# coding=utf-8
# ======================================
# File:     test_parameter.py
# Author:   Jackie PENG
# Contact:  jackie.pengzhao@gmail.com
# Created:  2025-08-12
# Desc:
#   Unittest for all functionalities of
# Parameter class, including blending
# functions.
# ======================================

import unittest
import numpy as np
import math

from qteasy.parameter import Parameter
from qteasy.blender import (
    blender_parser,
    signal_blend,
    human_blender,
    _exp_to_token,
)


class TestParameter(unittest.TestCase):
    def test_parameter(self):
        p = Parameter((0, 10),name='par1')
        self.assertEqual(p.name, 'par1')
        self.assertEqual(p.par_type, 'int')
        self.assertEqual(p.par_range, (0, 10))
        self.assertEqual(p.count, 11)
        self.assertEqual(p.size, 11)
        self.assertEqual(p.shape, None)
        self.assertEqual(p.dim, 0)
        self.assertEqual(p.array_size, 1)
        self.assertEqual(p.upper_bound, 10)
        self.assertEqual(p.lower_bound, 0)
        self.assertEqual(p.value, None)
        self.assertTrue(np.allclose(p.gen_values(11, 'int'), [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10]))
        self.assertRaises(ValueError, p.gen_values, 0.5, 'int')
        extracted = p.gen_values(8, 'rand')
        self.assertEqual(len(extracted), 8)
        self.assertTrue(all([(0 <= item <= 10) for item in extracted]))
        self.assertTrue(all([item in p for item in extracted]))
        self.assertIsNone(p.value)
        p.set_value(7)
        self.assertEqual(p.value, 7)
        p.value=8
        self.assertEqual(p.value, 8)
        self.assertEqual(list(p.enum_values()), [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10])
        self.assertEqual(str(p), "Parameter((0, 10), 'int')")

        self.assertRaises(TypeError, p.set_value, 'wrong_type')
        self.assertRaises(ValueError, p.set_value, 7.5)
        self.assertRaises(ValueError, p.set_value, 11)

        p = Parameter((0.0, 1.0), name='par2')
        self.assertEqual(p.name, 'par2')
        self.assertEqual(p.par_type, 'float')
        self.assertEqual(p.par_range, (0.0, 1.0))
        self.assertEqual(p.count, np.inf)
        self.assertEqual(p.size, 1.0)
        self.assertEqual(p.shape, None)
        self.assertEqual(p.dim, 0)
        self.assertEqual(p.array_size, 1)
        self.assertEqual(p.upper_bound, 1.0)
        self.assertEqual(p.lower_bound, 0.0)
        self.assertEqual(p.ubound, 1.0)
        self.assertEqual(p.lbound, 0.0)
        print(p.gen_values(2, 'interval'))
        self.assertTrue(np.allclose(p.gen_values(3, 'interval'), [0.0, 0.5, 1.0]))
        gen_values = p.gen_values(8, 'rand')
        self.assertEqual(len(gen_values), 8)
        self.assertTrue(all([(0.0 <= item <= 1.0) for item in gen_values]))
        self.assertIsNone(p.value)
        p.set_value(0.5)
        self.assertEqual(p.value, 0.5)
        p.value = 0.75
        self.assertEqual(p.get_value(), 0.75)
        self.assertEqual(str(p), "Parameter((0.0, 1.0), 'float')")

        self.assertRaises(TypeError, p.set_value, 'wrong_type')
        self.assertRaises(ValueError, p.set_value, 1.5)

        self.assertRaises(TypeError, p.enum_values)

        p = Parameter(('a', 'b'), name='par3')
        self.assertEqual(p.name, 'par3')
        self.assertEqual(p.par_type, 'enum')
        self.assertEqual(p.par_range, ('a', 'b'))
        self.assertEqual(p.count, 2)
        self.assertEqual(p.size, 2)
        self.assertEqual(p.shape, None)
        self.assertEqual(p.dim, 0)
        self.assertEqual(p.array_size, 1)
        self.assertEqual(p.ubound, 'b')
        self.assertEqual(p.lbound, 'a')
        self.assertEqual(p.upper_bound, 'b')
        self.assertEqual(p.lower_bound, 'a')
        self.assertEqual(list(p.gen_values(2, 'int')), ['a', 'b'])
        values = list(p.gen_values(3, 'rand'))
        self.assertEqual(len(values), 3)
        self.assertTrue(all([item in ('a', 'b') for item in values]))
        self.assertTrue(all([item in p for item in values]))
        self.assertEqual(list(p.enum_values()), ['a', 'b'])
        self.assertTrue('a' in p)

        # test parameter with array values
        p = Parameter((0, 10), name='arr_par', par_type='array[4, 5]')
        self.assertEqual(p.name, 'arr_par')
        self.assertEqual(p.par_type, 'int_array')
        self.assertEqual(p.par_range, (0, 10))
        self.assertEqual(p.shape, (4, 5))
        self.assertEqual(p.dim, 2)
        self.assertEqual(p.array_size, 20)
        self.assertEqual(p.count, 8667208279016151025)
        self.assertEqual(p.size, 8667208279016151025)
        self.assertEqual(p.ubound, 10)
        self.assertEqual(p.lbound, 0)
        self.assertEqual(p.upper_bound, 10)
        self.assertEqual(p.lower_bound, 0)
        for val in p.gen_values(5, 'int'):
            self.assertEqual(val.shape, (4, 5))
            self.assertTrue(np.all((0 <= val) & (val <= 10)))
            self.assertTrue(val in p)
        for val in p.gen_values(5, 'rand'):
            self.assertEqual(val.shape, (4, 5))
            self.assertTrue(np.all((0 <= val) & (val <= 10)))
            self.assertTrue(val in p)

        value1 = np.random.randint(0, 11, size=(4, 5))
        p.set_value(value1)
        self.assertTrue(np.array_equal(p.value, value1))
        self.assertTrue(value1 in p)
        value2 = np.random.randint(10, 15, size=(4, 5))
        self.assertFalse(value2 in p)
        self.assertRaises(ValueError, p.set_value, value2)
        value3 = np.random.randint(0, 11, size=(4, 4))
        self.assertFalse(value3 in p)
        self.assertFalse(5 in p)
        self.assertFalse('a' in p)
        self.assertRaises(ValueError, p.set_value, value3)

        self.assertRaises(ArithmeticError, p.enum_values)

        p = Parameter((0.0, 1.5), name='float_arr_par', par_type='array[3, 5]')
        self.assertEqual(p.par_type, 'float_array')
        self.assertEqual(p.name, 'float_arr_par')
        self.assertEqual(p.par_range, (0.0, 1.5))
        self.assertEqual(p.shape, (3, 5))
        self.assertEqual(p.dim, 2)
        self.assertEqual(p.array_size, 15)
        self.assertEqual(p.count, np.inf)
        self.assertAlmostEqual(p.size, 437.8938903808594)

        for val in p.gen_values(5, 'int'):
            self.assertEqual(val.shape, (3, 5))
            print(val)
            self.assertTrue(np.all((0 <= val) & (val <= 1.5)))
            self.assertTrue(val in p)
        for val in p.gen_values(5, 'rand'):
            self.assertEqual(val.shape, (3, 5))
            print(val)
            self.assertTrue(np.all((0 <= val) & (val <= 1.5)))
            self.assertTrue(val in p)

        p = Parameter((1, 10), name='int_arr_par', par_type='int[3, 5]')
        self.assertEqual(p.par_type, 'int_array')
        p = Parameter(10, name='int_arr_par', par_type='int[3, 5]')
        self.assertEqual(p.par_type, 'int_array')
        p = Parameter((0, 10), name='float_arr_par', par_type='float[3, 5]')
        self.assertEqual(p.par_type, 'float_array')
        p = Parameter(1.5, name='float_arr_par', par_type='float[3, 5, 5]')
        self.assertEqual(p.par_type, 'float_array')
        self.assertEqual(p.shape, (3, 5, 5))
        p = Parameter((1, 10), name='int_arr_par', par_type='int[3,]')
        self.assertEqual(p.par_type, 'int_array')

        self.assertRaises(ValueError, Parameter, 15, par_type='int[]')

        p = Parameter((1, 2, 3), name='par4')
        self.assertEqual(p.name, 'par4')
        self.assertEqual(p.par_type, 'enum')
        self.assertEqual(str(p), "Parameter((1, 2, 3), 'enum')")

        p = Parameter(3, name='par5')
        self.assertEqual(p.name, 'par5')
        self.assertEqual(p.par_type, 'int')
        self.assertEqual(p.par_range, (0, 3))

        p = Parameter(3.5, name='par6')
        self.assertEqual(p.name, 'par6')
        self.assertEqual(p.par_type, 'float')
        self.assertEqual(p.par_range, (0.0, 3.5))

        p = Parameter('a', name='par7')
        self.assertEqual(p.name, 'par7')
        self.assertEqual(p.par_type, 'enum')
        self.assertEqual(p.par_range, ('a',))

        p = Parameter(5, par_type='discr')
        self.assertEqual(p.name, '')
        self.assertEqual(p.par_type, 'int')

        self.assertRaises(ValueError, Parameter, 5, par_type='wrong_type')

        p = Parameter((0, 10), name='par8', par_type='int', value=5)
        self.assertEqual(p.name, 'par8')
        self.assertEqual(p.par_type, 'int')
        self.assertEqual(p.value, 5)

        p = Parameter((0.0, 8.0), name='par9', par_type='int')
        self.assertEqual(p.name, 'par9')
        self.assertEqual(p.par_type, 'int')

        self.assertRaises(ValueError, Parameter, (0.0, 8.0), name='par10', par_type='int', value=9.5)

        p = Parameter((10, -10), name='par10', par_type='int')
        self.assertEqual(p.name, 'par10')
        self.assertEqual(p.par_type, 'int')
        self.assertEqual(p.par_range, (-10, 10))
        self.assertEqual(p.count, 21)
        self.assertEqual(p.size, 21)
        self.assertEqual(p.ubound, 10)
        values = p.gen_values(5, 'interval')
        self.assertTrue(np.allclose(values, [-10, -5, 0, 5, 10]))

        p = Parameter((-10.5, 10.5), par_type='int')
        self.assertEqual(p.name, '')
        self.assertEqual(p.par_type, 'int')
        self.assertEqual(p.par_range, (-10, 10))

        # test other valid parameter initializations
        p = Parameter((10, '20'), name='par16', par_type='int')
        p = Parameter(('20',), name='par16', par_type='int')
        p = Parameter('20', name='par16')
        p = Parameter('a')
        p = Parameter('a, b, b, c')
        p = Parameter(['a', 'b', 'b', 'c'])

        # test wrong parameter initialization
        self.assertRaises(ValueError, Parameter, (10, -10), name='par11', par_type='int', value=15)
        self.assertRaises(ValueError, Parameter, (10, -10), name='par12', par_type='arr')
        self.assertRaises(ValueError, Parameter, (10, -10), name='par13', par_type='enum[2,3,4]')
        self.assertRaises(ValueError, Parameter, (10, -10), name='par14', par_type='int[2,3][2,3,4]')
        self.assertRaises(ValueError, Parameter, (10, -10), name='par15', par_type='wrong type')
        self.assertRaises(ValueError, Parameter, ('a', 'a20', 4, 4, 5), name='par16', par_type='int')
        self.assertRaises(ValueError, Parameter, (10, -10), name='par12', par_type='arr[-2, 3]')
        self.assertRaises(ValueError, Parameter, (10, -10), name='par12', par_type='arr[-2, "a"]')
        self.assertRaises(ValueError, Parameter, (10, -10), name='par12', par_type='arr[-2, 3.35]')

    def test_gen_values(self):
        p = Parameter((0, 10), name='par11', par_type='int')
        values = p.gen_values(6, 'interval')
        self.assertTrue(np.allclose(values, [0, 2, 4, 6, 8, 10]))
        values = p.gen_values(5, 'rand')
        self.assertEqual(len(values), 5)
        self.assertTrue(all([(0 <= item <= 10) for item in values]))

        p = Parameter((0.0, 10.0), name='par12', par_type='float')
        values = p.gen_values(5, 'interval')
        self.assertTrue(np.allclose(values, [0.0, 2.5, 5.0, 7.5, 10]))
        values = p.gen_values(5, 'rand')
        self.assertEqual(len(values), 5)
        self.assertTrue(all([(0.0 <= item <= 10.0) for item in values]))

        p = Parameter(('a', 'b', 'c', 'd', 'e', 'f', 'g'), name='par13', par_type='enum')
        values = p.gen_values(3, how='int')
        self.assertEqual(values, ['a', 'd', 'g'])
        values = p.gen_values(3, how='rand')
        self.assertEqual(len(values), 3)
        self.assertTrue(all([item in ('a', 'b', 'c', 'd', 'e', 'f', 'g') for item in values]))

        self.assertRaises(ValueError, p.gen_values, 0, 'int')
        self.assertRaises(ValueError, p.gen_values, -1, 'int')
        self.assertRaises(ValueError, p.gen_values, 0.35, 'int')
        self.assertRaises(ValueError, p.gen_values, 0, 'rand')
        self.assertRaises(ValueError, p.gen_values, -1, 'rand')
        self.assertRaises(ValueError, p.gen_values, 0.35, 'rand')
        self.assertRaises(TypeError, p.gen_values, 5, 15)
        self.assertRaises(KeyError, p.gen_values, 5, 'wrong_how')

    def test_get_and_set_values(self):
        p = Parameter((0, 10), name='par14', par_type='int')
        self.assertIsNone(p.value)
        p.set_value(5)
        self.assertEqual(p.value, 5)
        p.value = 7
        self.assertEqual(p.value, 7)
        self.assertEqual(p.get_value(), 7)

        self.assertRaises(TypeError, p.set_value, 'wrong_type')
        self.assertRaises(ValueError, p.set_value, 11)

        p = Parameter((0.0, 10.0), name='par15', par_type='float')
        self.assertIsNone(p.value)
        p.set_value(5.5)
        self.assertEqual(p.value, 5.5)
        p.value = 7.25
        self.assertEqual(p.value, 7.25)
        self.assertEqual(p.get_value(), 7.25)

        self.assertRaises(TypeError, p.set_value, 'wrong_type')
        self.assertRaises(ValueError, p.set_value, 11.5)

        p = Parameter(('a', 'b', 'c'), name='par16', par_type='enum')
        self.assertIsNone(p.value)
        p.set_value('b')
        self.assertEqual(p.value, 'b')
        p.value = 'c'
        self.assertEqual(p.value, 'c')
        self.assertEqual(p.get_value(), 'c')

        self.assertRaises(ValueError, p.set_value, 5)
        self.assertRaises(ValueError, p.set_value, 'd')

        p = Parameter((0, 10), name='par17', par_type='array[2,3]')
        self.assertIsNone(p.value)
        value1 = np.random.randint(0, 11, size=(2, 3))
        p.set_value(value1)
        self.assertTrue(np.array_equal(p.value, value1))
        value2 = np.random.randint(0, 11, size=(2, 3))
        p.value = value2
        self.assertTrue(np.array_equal(p.value, value2))
        self.assertTrue(np.array_equal(p.get_value(), value2))

        value3 = np.random.randint(11, 15, size=(2, 3))
        self.assertRaises(ValueError, p.set_value, value3)
        value4 = np.random.randint(0, 11, size=(2, 2))
        self.assertRaises(ValueError, p.set_value, value4)

    def test_update_range(self):
        """ test method update_par_range with various parameter types,
        test with valid and invalid inputs"""
        p = Parameter((0, 10), name='par18', par_type='int')
        p.update_par_range((5, 15))
        self.assertEqual(p.par_range, (5, 15))
        self.assertEqual(p.lower_bound, 5)
        self.assertEqual(p.upper_bound, 15)
        self.assertEqual(p.count, 11)
        self.assertEqual(p.size, 11)

        p.update_par_range((15, 5))
        self.assertEqual(p.par_range, (5, 15))
        self.assertEqual(p.lower_bound, 5)
        self.assertEqual(p.upper_bound, 15)
        self.assertEqual(p.count, 11)
        self.assertEqual(p.size, 11)

        self.assertRaises(TypeError, p.update_par_range, (5.5, 15))
        self.assertRaises(TypeError, p.update_par_range, (5, '15'))
        self.assertRaises(TypeError, p.update_par_range, 'wrong_type')

        p = Parameter((0.0, 10.0), name='par19', par_type='float')
        p.update_par_range((5.0, 15.0))
        self.assertEqual(p.par_range, (5.0, 15.0))
        self.assertEqual(p.lower_bound, 5.0)
        self.assertEqual(p.upper_bound, 15.0)
        self.assertEqual(p.count, np.inf)
        self.assertEqual(p.size, 10.0)

        self.assertRaises(ValueError, p.update_par_range, (5.0, 15.0, 16.0))
        self.assertRaises(TypeError, p.update_par_range, (5.0, '15.0'))
        self.assertRaises(TypeError, p.update_par_range, 'wrong_type')

        p = Parameter(('a', 'b', 'c'), name='par20', par_type='enum')
        p.update_par_range(('b', 'c', 'd', 'e'))
        self.assertEqual(p.par_range, ('b', 'c', 'd', 'e'))
        self.assertEqual(p.lower_bound, 'b')
        self.assertEqual(p.upper_bound, 'e')
        self.assertEqual(p.count, 4)
        self.assertEqual(p.size, 4)

        self.assertRaises(TypeError, p.update_par_range, {'a', 'b', 'c'})
        self.assertRaises(TypeError, p.update_par_range, 'wrong_type')

    def test_signal_blend(self):
        self.assertEqual(blender_parser('s0 & 1'), ['&', '1', 's0'])
        self.assertEqual(blender_parser('s0 or 1'), ['or', '1', 's0'])
        self.assertEqual(blender_parser('s0 & s1 | s2'), ['|', 's2', '&', 's1', 's0'])
        blender = blender_parser('s0 & s1 | s2')
        self.assertEqual(signal_blend([1, 1, 1], blender), 1)
        self.assertEqual(signal_blend([1, 0, 1], blender), 1)
        self.assertEqual(signal_blend([1, 1, 0], blender), 1)
        self.assertEqual(signal_blend([0, 1, 1], blender), 1)
        self.assertEqual(signal_blend([0, 0, 1], blender), 1)
        self.assertEqual(signal_blend([1, 0, 0], blender), 0)
        self.assertEqual(signal_blend([0, 1, 0], blender), 0)
        self.assertEqual(signal_blend([0, 0, 0], blender), 0)
        # parse: '0 & ( 1 | 2 )'
        self.assertEqual(blender_parser('s0 & ( s1 | s2 )'), ['&', '|', 's2', 's1', 's0'])
        blender = blender_parser('s0 & ( s1 | s2 )')
        self.assertEqual(signal_blend([1, 1, 1], blender), 1)
        self.assertEqual(signal_blend([1, 0, 1], blender), 1)
        self.assertEqual(signal_blend([1, 1, 0], blender), 1)
        self.assertEqual(signal_blend([0, 1, 1], blender), 0)
        self.assertEqual(signal_blend([0, 0, 1], blender), 0)
        self.assertEqual(signal_blend([1, 0, 0], blender), 0)
        self.assertEqual(signal_blend([0, 1, 0], blender), 0)
        self.assertEqual(signal_blend([0, 0, 0], blender), 0)
        # parse: '(1-2) and 3 + ~0'
        self.assertEqual(blender_parser('(s1-s2)/s3 + s0'), ['+', 's0', '/', 's3', '-', 's2', 's1'])
        blender = blender_parser('(s1-s2)/s3 + s0')
        self.assertEqual(signal_blend([5, 9, 1, 4], blender), 7)
        # parse: '(1-2)/3 + 0'
        self.assertEqual(blender_parser('(s1-s2)/s3 + s0'), ['+', 's0', '/', 's3', '-', 's2', 's1'])
        blender = blender_parser('(s1-s2)/s3 + s0')
        self.assertEqual(signal_blend([5, 9, 1, 4], blender), 7)
        # pars: '(0*1/2*(3+4))+5*(6+7)-8'
        self.assertEqual(blender_parser('(s0*s1/s2*(s3+s4))+s5*(s6+s7)-s8'),
                         ['-', 's8', '+', '*', '+', 's7', 's6', 's5', '*', '+', 's4', 's3', '/', 's2', '*', 's1', 's0'])
        blender = blender_parser('(s0*s1/s2*(s3+s4))+s5*(s6+s7)-s8')
        self.assertEqual(signal_blend([1, 1, 1, 1, 1, 1, 1, 1, 1], blender), 3)
        self.assertEqual(signal_blend([2, 1, 4, 3, 5, 5, 2, 2, 10], blender), 14)
        # parse: '0/max(2,1,3 + 5)+4'
        self.assertEqual(blender_parser('s0/max(s2,s1,s3 + s5)+s4'),
                         ['+', 's4', '/', 'max(3)', '+', 's5', 's3', 's1', 's2', 's0'])
        blender = blender_parser('s0/max(s2,s1,s3 + 5)+s4')
        self.assertEqual(signal_blend([8.0, 4, 3, 5.0, 0.125, 5], blender), 0.925)
        self.assertEqual(signal_blend([2, 1, 4, 3, 5, 5, 2, 2, 10], blender), 5.25)

        print('speed test')
        import time
        st = time.time()
        blender = blender_parser('s0+max(s1,s2,(s3+s4)*s5, max(s6, (s7+s8)*s9), s10-s11) * (s12+s13)')
        res = []
        for i in range(10000):
            res = signal_blend([1, 1, 2, 3, 4, 5, 3, 4, 5, 6, 7, 8, 2, 3], blender)
        et = time.time()
        print(f'total time for RPN processing: {et - st}, got result: {res}')

        blender = blender_parser("s0 + s1 * s2")
        self.assertEqual(signal_blend([1, 2, 3], blender), 7)
        blender = blender_parser("(s0 + s1) * s2")
        self.assertEqual(signal_blend([1, 2, 3], blender), 9)
        blender = blender_parser("(s0+s1) * s2")
        self.assertEqual(signal_blend([1, 2, 3], blender), 9)
        blender = blender_parser("(s0 + s1)   * s2")
        self.assertEqual(signal_blend([1, 2, 3], blender), 9)
        blender = blender_parser("(s0-s1)/s2 + s3")
        print(f'RPN of notation: "(s0-s1)/s2 + s3" is:\n'
              f'{" ".join(blender[::-1])}')
        self.assertAlmostEqual(signal_blend([1, 2, 3, 0.0], blender), -0.33333333)
        blender = blender_parser("-(s0-s1)/s2 + s3")
        print(f'RPN of notation: "-(s0-s1)/s2 + s3" is:\n'
              f'{" ".join(blender[::-1])}')
        self.assertAlmostEqual(signal_blend([1, 2, 3, 0.0], blender), 0.33333333)
        blender = blender_parser("~(0-1)/s2 + s3")
        print(f'RPN of notation: "~(s0-s1)/s2 + s3" is:\n'
              f'{" ".join(blender[::-1])}')
        self.assertAlmostEqual(signal_blend([1, 2, 3, 0.0], blender), 0.33333333)
        blender = blender_parser("s0 + s1 / s2")
        print(f'RPN of notation: "0 + 1 / 2" is:\n'
              f'{" ".join(blender[::-1])}')
        self.assertAlmostEqual(signal_blend([1, math.pi, 4], blender), 1.78539816)
        blender = blender_parser("(s0 + s1) / s2")
        print(f'RPN of notation: "(0 + 1) / 2" is:\n'
              f'{" ".join(blender[::-1])}')
        self.assertEqual(signal_blend([1, 2, 3], blender), 1)
        blender = blender_parser("(s0 + s1 * s2) / s3")
        print(f'RPN of notation: "(0 + 1 * 2) / 3" is:\n'
              f'{" ".join(blender[::-1])}')
        self.assertAlmostEqual(signal_blend([3, math.e, 10, 10], blender), 3.0182818284590454)
        blender = blender_parser("s0 / s1 * s2")
        print(f'RPN of notation: "0 / 1 * 2" is:\n'
              f'{" ".join(blender[::-1])}')
        self.assertEqual(signal_blend([1, 3, 6], blender), 2)
        blender = blender_parser("(s0 - s1 + s2) * s4")
        print(f'RPN of notation: "(0 - 1 + 2) * 4" is:\n'
              f'{" ".join(blender[::-1])}')
        self.assertAlmostEqual(signal_blend([1, 1, -1, np.nan, math.pi], blender), -3.141592653589793)
        blender = blender_parser("s0 * s1")
        print(f'RPN of notation: "0 * 1" is:\n'
              f'{" ".join(blender[::-1])}')
        self.assertAlmostEqual(signal_blend([math.pi, math.e], blender), 8.539734222673566)

        blender = blender_parser('abs(s3-sqrt(s2) /  cos(s1))')
        print(f'RPN of notation: "abs(3-sqrt(2) /  cos(1))" is:\n'
              f'{" ".join(blender[::-1])}')
        self.assertEqual(blender, ['abs(1)', '-', '/', 'cos(1)', 's1', 'sqrt(1)', 's2', 's3'])
        blender = blender_parser('s0/max(s2,s1,s3 + s5)+s4')
        print(f'RPN of notation: "0/max(2,1,3 + 5)+4" is:\n'
              f'{" ".join(blender[::-1])}')
        self.assertEqual(blender, ['+', 's4', '/', 'max(3)', '+', 's5', 's3', 's1', 's2', 's0'])

        blender = blender_parser('s1 + sum(s1,s2,s3+s3, sum(s1, s2) + s3) *s5')
        print(f'RPN of notation: "1 + sum(1,2,3+3, sum(1, 2) + 3) *5" is:\n'
              f'{" ".join(blender[::-1])}')
        self.assertEqual(blender, ['+', '*', 's5', 'sum(4)', '+', 's3', 'sum(2)', 's2', 's1',
                                   '+', 's3', 's3', 's2', 's1', 's1'])
        blender = blender_parser('s1+sum(1,2,(s3+s5)*s4, sum(s3, (4+s5)*s6), s7-s8) * (s2+s3)')
        print(f'RPN of notation: "1+sum(1,2,(3+5)*4, sum(3, (4+5)*6), 7-8) * (2+3)" is:\n'
              f'{" ".join(blender[::-1])}')
        self.assertEqual(blender, ['+', '*', '+', 's3', 's2', 'sum(5)', '-', 's8', 's7',
                                   'sum(2)', '*', 's6', '+', 's5', '4', 's3', '*', 's4',
                                   '+', 's5', 's3', '2', '1', 's1'])

    def test_tokenizer(self):
        self.assertListEqual(_exp_to_token('s1+s1'),
                             ['s1', '+', 's1'])
        print(_exp_to_token('s1+s1'))
        self.assertListEqual(_exp_to_token('1+1'),
                             ['1', '+', '1'])
        print(_exp_to_token('1+1'))
        self.assertListEqual(_exp_to_token('1 & 1'),
                             ['1', '&', '1'])
        print(_exp_to_token('1&1'))
        self.assertListEqual(_exp_to_token('1 and 1'),
                             ['1', 'and', '1'])
        print(_exp_to_token('1 and 1'))
        self.assertListEqual(_exp_to_token('-1 and 1'),
                             ['-1', 'and', '1'])
        print(_exp_to_token('s0 and s1'))
        self.assertListEqual(_exp_to_token('s0 or s1'),
                             ['s0', 'or', 's1'])
        print(_exp_to_token('1 and 1'))
        self.assertListEqual(_exp_to_token('1 or 1'),
                             ['1', 'or', '1'])
        print(_exp_to_token('1 or 1'))
        self.assertListEqual(_exp_to_token('(1 - 1 + -1) * pi'),
                             ['(', '1', '-', '1', '+', '-1', ')', '*', 'pi'])
        print(_exp_to_token('(1 - 1 + -1) * pi'))
        self.assertListEqual(_exp_to_token('(s1 - s1 + -1) * pi'),
                             ['(', 's1', '-', 's1', '+', '-1', ')', '*', 'pi'])
        print(_exp_to_token('(s1 - s1 + -1) * pi'))
        self.assertListEqual(_exp_to_token('abs(5-sqrt(2) /  cos(pi))'),
                             ['abs(', '5', '-', 'sqrt(', '2', ')', '/', 'cos(', 'pi', ')', ')'])
        print(_exp_to_token('abs(5-sqrt(2) /  cos(pi))'))
        self.assertListEqual(_exp_to_token('abs(s5-sqrt(s2) /  cos(pi))'),
                             ['abs(', 's5', '-', 'sqrt(', 's2', ')', '/', 'cos(', 'pi', ')', ')'])
        print(_exp_to_token('abs(s5-sqrt(s2) /  cos(pi))'))
        self.assertListEqual(_exp_to_token('sin(pi) + 2.14'),
                             ['sin(', 'pi', ')', '+', '2.14'])
        print(_exp_to_token('sin(pi) + 2.14'))
        self.assertListEqual(_exp_to_token('-sin(pi) + 2.14'),
                             ['-1', '*', 'sin(', 'pi', ')', '+', '2.14'])
        print(_exp_to_token('-sin(pi) + 2.14'))
        self.assertListEqual(_exp_to_token('(1-2)/3.0 + 0.0000'),
                             ['(', '1', '-', '2', ')', '/', '3.0', '+', '0.0000'])
        print(_exp_to_token('(1-2)/3.0 + 0.0000'))
        self.assertListEqual(_exp_to_token('-(1. + .2) * max(1, 3, 5)'),
                             ['-1', '*', '(', '1.', '+', '.2', ')', '*', 'max(', '1', ',', '3', ',', '5', ')'])
        print(_exp_to_token('-(1. + .2) * max(1, 3, 5)'))
        self.assertListEqual(_exp_to_token('(x + e * 10) / 10'),
                             ['(', 'x', '+', 'e', '*', '10', ')', '/', '10'])
        print(_exp_to_token('(x + e * 10) / 10'))
        self.assertListEqual(_exp_to_token('8.2/((-.1+abs3(3,4,5))*0.12)'),
                             ['8.2', '/', '(', '(', '-.1', '+', 'abs3(', '3', ',', '4', ',', '5', ')', ')', '*', '0.12',
                              ')'])
        print(_exp_to_token('8.2/((-.1+abs3(3,4,5))*0.12)'))
        self.assertListEqual(_exp_to_token('8.2/abs3(3,4,25.34 + 5)*0.12'),
                             ['8.2', '/', 'abs3(', '3', ',', '4', ',', '25.34', '+', '5', ')', '*', '0.12'])
        print(_exp_to_token('8.2/abs3(3,4,25.34 + 5)*0.12'))
        self.assertListEqual(_exp_to_token('abs(-1.14)+power(2, 3)and log(3.14)'),
                             ['abs(', '-1.14', ')', '+', 'power(', '2', ',', '3', ')', 'and', 'log(', '3.14', ')'])
        print(_exp_to_token('abs(-1.14)+power(2, 3)and log(3.14)'))
        self.assertListEqual(_exp_to_token('strength_1.25(0, 1, 2)'),
                             ['strength_1.25(', '0', ',', '1', ',', '2', ')'])
        print(_exp_to_token('strength_1.25(0, 1, 2)'))
        self.assertListEqual(_exp_to_token('avg_pos_3_1.25(0, 1,2)'),
                             ['avg_pos_3_1.25(', '0', ',', '1', ',', '2', ')'])
        print(_exp_to_token('avg_pos_3_1.25(0, 1,2)'))
        self.assertListEqual(_exp_to_token('clip_-1_1(pos_5_0.2(0, 1, 2, 3, 4))'),
                             ['clip_-1_1(', 'pos_5_0.2(', '0', ',', '1', ',', '2', ',', '3', ',', '4', ')', ')'])
        print(_exp_to_token('clip_-1_1(pos_5_0.2(0, 1, 2, 3, 4))'))

    def test_all_blending_funcs(self):
        """ 测试其他信号组合函数是否正常工作"""
        # 生成五个示例交易信号
        signal0 = np.array([[0.12, 0.35, 0.11],
                            [0.81, 0.22, 0.29],
                            [0.86, 0.47, 0.29],
                            [0.46, 0.81, 0.60],
                            [0.42, 0.55, 0.74]])
        signal1 = np.array([[0.94, 0.66, 0.69],
                            [0.85, 0.30, 0.65],
                            [0.87, 0.06, 0.24],
                            [0.73, 0.20, 0.19],
                            [0.43, 0.18, 0.44]])
        signal2 = np.array([[0.24, 0.81, 0.44],
                            [0.66, 0.92, 0.99],
                            [0.18, 0.17, 0.11],
                            [0.48, 0.57, 0.55],
                            [0.37, 0.66, 0.01]])
        signal3 = np.array([[0.92, 0.88, 0.16],
                            [0.89, 0.79, 0.27],
                            [0.48, 0.77, 0.20],
                            [0.43, 0.33, 0.25],
                            [0.90, 0.30, 0.49]])
        signal4 = np.array([[0.05, 0.17, 0.30],
                            [0.16, 0.62, 0.61],
                            [0.52, 0.83, 0.57],
                            [0.16, 0.36, 0.28],
                            [0.99, 0.57, 0.04]])
        # 将示例信号组合为交易信号组，与operator输出形式相同
        signals = [signal0, signal1, signal2, signal3, signal4]

        # 开始测试blender functions
        print('\ntest average functions')
        blender_exp = 'avg(s0, s1, s2, s3, s4)'
        blender = blender_parser(blender_exp)
        res = signal_blend(signals, blender)
        print(f'blended signals with blender "{blender_exp}" is \n{res}')
        target = np.array([[0.454, 0.574, 0.340],
                           [0.674, 0.570, 0.562],
                           [0.582, 0.460, 0.282],
                           [0.452, 0.454, 0.374],
                           [0.622, 0.452, 0.344]])

        hit = np.allclose(res, target)
        self.assertTrue(hit)

        print('\ntest comparison functions')
        blender_exp = 'combo(s0, s1, s2) + min(s0, s1,s2)-max(s2, s3, s4)'
        blender = blender_parser(blender_exp)
        res = signal_blend(signals, blender)
        print(f'blended signals with blender "{blender_exp}" is \n{res}')
        target = np.array([[0.50,  1.29,  0.91],
                           [2.09,  0.74,  1.23],
                           [1.57, -0.07,  0.18],
                           [1.65,  1.21,  0.98],
                           [0.60,  0.91,  0.71]])

        hit = np.allclose(res, target)
        self.assertTrue(hit)

        print('\ntest mathematical functions')
        blender_exp = 'abs(s0) + ceil(s1) * pow(s0, s1) + floor(s2+s3+s4) - exp(s3) and log(s4)'
        blender = blender_parser(blender_exp)
        res = signal_blend(signals, blender)
        print(f'blended signals with blender "{blender_exp}" is \n{res}')
        target = np.array([[8.77344165, 6.12214254, 1.74092762],
                           [7.10858499, 3.90823377, 2.38476921],
                           [3.79382222, 2.82813777, 1.71955085],
                           [4.84445021, 4.18981584, 4.14202468],
                           [3.13336769, 3.20675832, 6.87013820]])

        hit = np.allclose(res, target)
        self.assertTrue(hit)

        print('\ntest signal combination function strength')
        blender_exp = 'strength_1.35(s0, s1, s2)'
        blender = blender_parser(blender_exp)
        res = signal_blend(signals, blender)
        print(f'blended signals with blender "{blender_exp}" is \n{res}')
        target = np.array([[0., 1., 0.],
                           [1., 1., 1.],
                           [1., 0., 0.],
                           [1., 1., 0.],
                           [0., 1., 0.]])

        hit = np.allclose(res, target)
        self.assertTrue(hit)

        print('\ntest signal combination function position')
        blender_exp = 'pos_3_0.5(s0, s1, s2, s3, s4)'
        blender = blender_parser(blender_exp)
        res = signal_blend(signals, blender)
        print(f'blended signals with blender "{blender_exp}" is \n{res}')
        target = np.array([[0., 1., 0.],
                           [1., 1., 1.],
                           [1., 0., 0.],
                           [0., 0., 0.],
                           [0., 1., 0.]])

        hit = np.allclose(res, target)
        self.assertTrue(hit)

        print('\ntest signal combination function clip')
        blender_exp = 'clip_-1_0.8(pos_5_0.2(s0, s1, s2, s3, s4))'
        blender = blender_parser(blender_exp)
        res = signal_blend(signals, blender)
        print(f'blended signals with blender "{blender_exp}" is \n{res}')
        target = np.array([[0.0, 0.0, 0.0],
                           [0.0, 0.8, 0.8],
                           [0.0, 0.0, 0.0],
                           [0.0, 0.8, 0.0],
                           [0.8, 0.0, 0.0]])

        hit = np.allclose(res, target)
        self.assertTrue(hit)

        print('\ntest signal combination function avg position')
        blender_exp = 'avgpos_3_0.5(s0, s1, s2, s3, s4)'
        blender = blender_parser(blender_exp)
        res = signal_blend(signals, blender)
        print(f'blended signals with blender "{blender_exp}" is \n{res}')
        target = np.array([[0.000, 0.574, 0.000],
                           [0.674, 0.570, 0.562],
                           [0.582, 0.000, 0.000],
                           [0.000, 0.000, 0.000],
                           [0.000, 0.452, 0.000]])

        hit = np.allclose(res, target)
        self.assertTrue(hit)

        print('\ntest signal combination function unify')
        blender_exp = 'unify(avgpos_3_0.5(s0, s1, s2, s3, s4))'
        blender = blender_parser(blender_exp)
        res = signal_blend(signals, blender)
        print(f'blended signals with blender "{blender_exp}" is \n{res}')
        target = np.array([[0.00000000, 1.00000000, 0.00000000],
                           [0.37320044, 0.31561462, 0.31118494],
                           [1.00000000, 0.00000000, 0.00000000],
                           [0.00000000, 0.00000000, 0.00000000],
                           [0.00000000, 1.00000000, 0.00000000]])

        hit = np.allclose(res, target)
        self.assertTrue(hit)

        print('\ntest signal combination function with pure numbers')
        blender_exp = 'avgpos_3_0.5(s0, 1.5*s1, 2*s2, 0.5*s3, 2+s4)'
        blender = blender_parser(blender_exp)
        res = signal_blend(signals, blender)
        print(f'blended signals with blender "{blender_exp}" is \n{res}')
        target = np.array([[0.000, 1.114, 0.881],
                           [1.202, 0.000, 1.198],
                           [1.057, 0.000, 0.000],
                           [0.978, 0.955, 0.878],
                           [1.049, 0.972, 0.741]])

        hit = np.allclose(res, target)
        self.assertTrue(hit)

        # test human_blender function:
        print('\ntest human_blender function')
        strategy_ids = ['MACD', 'DMA', 'CROSSLINE', 'TRIX', 'KDJ']
        blender_exp = 'avgpos_3_0.5(s0, s1, s2, s3, s4)'
        res = human_blender(blender_exp, strategy_ids)
        print(f'blended signals with blender "{blender_exp}" is \n{res}')
        self.assertEqual(res, 'avgpos_3_0.5(MACD, DMA, CROSSLINE, TRIX, KDJ)')

        blender_exp = 'max(s0, s1, s2)+s3*s4'
        res = human_blender(blender_exp, strategy_ids)
        print(f'blended signals with blender "{blender_exp}" is \n{res}')
        self.assertEqual(res, 'max(MACD, DMA, CROSSLINE) + TRIX * KDJ')

        blender_exp = 'max(s0, s1/s0)+s1and0.5*s4'
        res = human_blender(blender_exp, strategy_ids)
        print(f'blended signals with blender "{blender_exp}" is \n{res}')
        self.assertEqual(res, 'max(MACD, DMA / MACD) + DMA and 0.5 * KDJ')

        blender_exp = 'max(s0, s1/s0)+s1^0.5*s6'
        self.assertRaises(IndexError, human_blender, blender_exp, strategy_ids)

        # false case
        self.assertEqual(human_blender(None, strategy_ids), '')
        self.assertRaises(TypeError, human_blender, 123, strategy_ids)
        self.assertRaises(TypeError, human_blender, blender_exp, None)

    def test_parameter_copy(self):
        p1 = Parameter((0, 10), name='par21', par_type='int', value=5)
        p2 = p1.copy()
        self.assertIsNot(p1, p2)
        self.assertEqual(p1.name, p2.name)
        self.assertEqual(p1.par_type, p2.par_type)
        self.assertEqual(p1.par_range, p2.par_range)
        self.assertEqual(p1.value, p2.value)

        p1 = Parameter((0.0, 10.0), name='par22', par_type='float', value=5.5)
        p2 = p1.copy()
        self.assertIsNot(p1, p2)
        self.assertEqual(p1.name, p2.name)
        self.assertEqual(p1.par_type, p2.par_type)
        self.assertEqual(p1.par_range, p2.par_range)
        self.assertEqual(p1.value, p2.value)

        p1 = Parameter(('a', 'b', 'c'), name='par23', par_type='enum', value='b')
        p2 = p1.copy()
        self.assertIsNot(p1, p2)
        self.assertEqual(p1.name, p2.name)
        self.assertEqual(p1.par_type, p2.par_type)
        self.assertEqual(p1.par_range, p2.par_range)
        self.assertEqual(p1.value, p2.value)

        p1 = Parameter((0, 10), name='par24', par_type='array[2,3]', value=np.random.randint(0, 11, size=(2, 3)))
        p2 = p1.copy()
        self.assertIsNot(p1, p2)
        self.assertEqual(p1.name, p2.name)
        self.assertEqual(p1.par_type, p2.par_type)
        self.assertEqual(p1.par_range, p2.par_range)
        self.assertTrue(np.array_equal(p1.value, p2.value))


if __name__ == '__main__':
    unittest.main()


# coding=utf-8
# ======================================
# File:     test_utilityfuncs.py
# Author:   Jackie PENG
# Contact:  jackie.pengzhao@gmail.com
# Created:  2020-02-12
# Desc:
#   Unittest for all qteasy utility
#   functionalities.
# ======================================
import unittest

import pandas as pd
import numpy as np
import datetime
import logging

from qteasy.utilfuncs import list_to_str_format, regulate_date_format, sec_to_duration, str_to_list
from qteasy.utilfuncs import maybe_trade_day, is_market_trade_day, prev_trade_day, next_trade_day
from qteasy.utilfuncs import next_market_trade_day, unify, list_or_slice, labels_to_dict, retry
from qteasy.utilfuncs import weekday_name, nearest_market_trade_day, is_number_like, list_truncate, input_to_list
from qteasy.utilfuncs import match_ts_code, _lev_ratio, _partial_lev_ratio, _wildcard_match, rolling_window
from qteasy.utilfuncs import reindent, adjust_string_length, is_float_like, is_integer_like
from qteasy.utilfuncs import is_cn_stock_symbol_like, is_complete_cn_stock_symbol_like
from qteasy.utilfuncs import sanitize_filename


class RetryableError(Exception):
    """ 用于retry测试的自定义Error Type"""
    pass


class AnotherRetryableError(Exception):
    """ 用于retry测试的自定义Error Type"""
    pass


class UnexpectedError(Exception):
    """ 用于retry测试的自定义Error Type"""
    pass


class TestUtilityFuncs(unittest.TestCase):
    def setUp(self):
        pass

    def test_input_to_list(self):
        print('Testing input_to_list() function')
        input_str = 'first'
        self.assertEqual(input_to_list(input_str, 3), ['first', 'first', 'first'])
        self.assertEqual(input_to_list(input_str, 4), ['first', 'first', 'first', 'first'])
        self.assertEqual(input_to_list(input_str, 2, None), ['first', 'first'])
        input_list = ['first', 'second']
        self.assertEqual(input_to_list(input_list, 3), ['first', 'second', None])
        self.assertEqual(input_to_list(input_list, 4, 'padder'), ['first', 'second', 'padder', 'padder'])
        self.assertEqual(input_to_list(input_list, 1), ['first', 'second'])
        self.assertEqual(input_to_list(input_list, -5), ['first', 'second'])

    def test_unify(self):
        n1 = 2
        self.assertEqual(unify(n1), 2)
        n2 = np.array([[1., 1., 1., 1.], [0., 1., 0., 1.]])
        n3 = np.array([[.25, .25, .25, .25], [0., .5, 0., .5]])
        self.assertTrue(np.allclose(unify(n2), n3))

    def test_time_string_format(self):
        print('Testing qt.time_string_format() function:')
        t = 3.14
        self.assertEqual(sec_to_duration(t), '3 sec 140.0 ms')
        self.assertEqual(sec_to_duration(t, estimation=True), 'about 3 sec')
        self.assertEqual(sec_to_duration(t, short_form=True), '3"140')
        self.assertEqual(sec_to_duration(t, estimation=True, short_form=True), '~3"')
        t = 300.14
        self.assertEqual(sec_to_duration(t), '5 min 140.0 ms')
        self.assertEqual(sec_to_duration(t, estimation=True), 'about 5 min')
        self.assertEqual(sec_to_duration(t, short_form=True), "5'140")
        self.assertEqual(sec_to_duration(t, estimation=True, short_form=True), "~5'")
        t = 7435.0014
        self.assertEqual(sec_to_duration(t), '2 hours 3 min 55 sec 1.4 ms')
        self.assertEqual(sec_to_duration(t, estimation=True), 'about 2 hours')
        self.assertEqual(sec_to_duration(t, short_form=True), "2H3'55\"001")
        self.assertEqual(sec_to_duration(t, estimation=True, short_form=True), "~2H")
        t = 86120.0509
        self.assertEqual(sec_to_duration(t), '23 hours 55 min 20 sec 50.9 ms')
        self.assertEqual(sec_to_duration(t, estimation=True), 'about 1 day')
        self.assertEqual(sec_to_duration(t, short_form=True), "23H55'20\"051")
        self.assertEqual(sec_to_duration(t, estimation=True, short_form=True), "~1D")
        t = 88425.0509
        self.assertEqual(sec_to_duration(t), '1 day 33 min 45 sec 50.9 ms')
        self.assertEqual(sec_to_duration(t, estimation=True), 'about 1 day 1 hour')
        self.assertEqual(sec_to_duration(t, short_form=True), "1D33'45\"051")
        self.assertEqual(sec_to_duration(t, estimation=True, short_form=True), "~1D1H")
        t = 288425.0509
        self.assertEqual(sec_to_duration(t), '3 days 8 hours 7 min 5 sec 50.9 ms')
        self.assertEqual(sec_to_duration(t, estimation=True), 'about 3 days 8 hours')
        self.assertEqual(sec_to_duration(t, short_form=True), "3D8H7'5\"051")
        self.assertEqual(sec_to_duration(t, estimation=True, short_form=True), "~3D8H")

    def test_str_to_list(self):
        self.assertEqual(str_to_list('a,b,c,d,e'), ['a', 'b', 'c', 'd', 'e'])
        self.assertEqual(str_to_list('a, b, c '), ['a', 'b', 'c'])
        self.assertEqual(str_to_list('a, b: c', sep_char=':'), ['a,b', 'c'])
        self.assertEqual(str_to_list('abc'), ['abc'])
        self.assertEqual(str_to_list(''), [])
        self.assertRaises(AssertionError, str_to_list, 123)

    def test_sanitize_filename(self):
        """文件名规范化：仅保留字母数字下划线连字符、空格转下划线、小写、去首点、无连续/首尾下划线、截断过长。"""
        print('\n[TestSanitizeFilename] 跨平台安全文件名规范化')
        self.assertEqual(sanitize_filename('Trade Log_2024.csv'), 'trade_log_2024.csv')
        self.assertEqual(sanitize_filename('  My File.Name  '), 'my_file.name')
        self.assertEqual(sanitize_filename('trade_log_MyOp_20240308_123456.csv'), 'trade_log_myop_20240308_123456.csv')
        self.assertEqual(sanitize_filename('live_log_1_User Name.log'), 'live_log_1_user_name.log')
        # 连续下划线合并、去掉开头和结尾的下划线
        self.assertEqual(sanitize_filename('file_name_with___under_score.csv'), 'file_name_with_under_score.csv')
        self.assertEqual(sanitize_filename('__file_name_starts_with_underscore.file'), 'file_name_starts_with_underscore.file')
        self.assertEqual(sanitize_filename('file_name_ends_with_underscore_.file'), 'file_name_ends_with_underscore.file')
        self.assertEqual(sanitize_filename('.hidden'), 'unnamed.hidden')
        self.assertEqual(sanitize_filename('UPPER.CSV'), 'upper.csv')
        self.assertEqual(sanitize_filename('a/b\\c*d?.txt'), 'abcd.txt')
        self.assertEqual(sanitize_filename(''), 'unnamed')
        self.assertEqual(sanitize_filename('  \t  '), 'unnamed')
        short = 'a' * 200 + '.csv'
        out = sanitize_filename(short, max_length=255)
        self.assertLessEqual(len(out), 255)
        self.assertTrue(out.endswith('.csv'))
        print('  sanitize_filename samples:', out[:50], '...', out[-10:])

    def test_list_or_slice(self):
        str_dict = {'close': 0, 'open': 1, 'high': 2, 'low': 3}
        self.assertEqual(list_or_slice(slice(1, 2, 1), str_dict), slice(1, 2, 1))
        self.assertEqual(list_or_slice('open', str_dict), [1])
        self.assertEqual(list(list_or_slice('close, high, low', str_dict)), [0, 2, 3])
        self.assertEqual(list(list_or_slice('close:high', str_dict)), [0, 1, 2])
        self.assertEqual(list(list_or_slice(['open'], str_dict)), [1])
        self.assertEqual(list(list_or_slice(['open', 'high'], str_dict)), [1, 2])
        self.assertEqual(list(list_or_slice(0, str_dict)), [0])
        self.assertEqual(list(list_or_slice([0, 2], str_dict)), [0, 2])
        self.assertEqual(list(list_or_slice([True, False, True, False], str_dict)), [0, 2])

    def test_labels_to_dict(self):
        target_list = [0, 1, 10, 100]
        target_dict = {'close': 0, 'open': 1, 'high': 2, 'low': 3}
        target_dict2 = {'close': 0, 'open': 2, 'high': 1, 'low': 3}
        self.assertEqual(labels_to_dict('close, open, high, low', target_list), target_dict)
        self.assertEqual(labels_to_dict(['close', 'open', 'high', 'low'], target_list), target_dict)
        self.assertEqual(labels_to_dict('close, high, open, low', target_list), target_dict2)
        self.assertEqual(labels_to_dict(['close', 'high', 'open', 'low'], target_list), target_dict2)

    def test_input_to_list(self):
        """ test util function input_to_list()"""
        self.assertEqual(input_to_list(5, 3), [5, 5, 5])
        self.assertEqual(input_to_list(5, 3, 0), [5, 5, 5])
        self.assertEqual(input_to_list([5], 3, 0), [5, 0, 0])
        self.assertEqual(input_to_list([5, 4], 3, 0), [5, 4, 0])

    def test_regulate_date_format(self):
        self.assertEqual(regulate_date_format('2019/11/06'), '2019-11-06')
        self.assertEqual(regulate_date_format('2019-11-06'), '2019-11-06')
        self.assertEqual(regulate_date_format('20191106'), '2019-11-06')
        self.assertEqual(regulate_date_format('191106'), '2006-11-19')
        self.assertEqual(regulate_date_format('830522'), '1983-05-22')
        self.assertEqual(regulate_date_format(datetime.datetime(2010, 3, 15)), '2010-03-15')
        self.assertEqual(regulate_date_format(datetime.datetime(2010, 3, 15, 12, 12, 23)), '2010-03-15 12:12:23')
        self.assertEqual(regulate_date_format('20191106 16:20:48'), '2019-11-06 16:20:48')
        self.assertEqual(regulate_date_format(pd.Timestamp('2010.03.15')), '2010-03-15')
        self.assertRaises(ValueError, regulate_date_format, 'abc')
        self.assertRaises(ValueError, regulate_date_format, '2019/13/43')

    def test_list_to_str_format(self):
        self.assertEqual(list_to_str_format(['close', 'open', 'high', 'low']),
                         'close,open,high,low')
        self.assertEqual(list_to_str_format(['letters', '  ', '123  4', 123, '   kk  l']),
                         'letters,,1234,kkl')
        self.assertEqual(list_to_str_format('a string input'),
                         'a,string,input')
        self.assertEqual(list_to_str_format('already,a,good,string'),
                         'already,a,good,string')
        self.assertRaises(AssertionError, list_to_str_format, 123)

    def test_is_trade_day(self):
        """test if the funcion maybe_trade_day() and is_market_trade_day() works properly
        """
        date_trade = '20210401'
        date_holiday = '20210102'
        date_weekend = '20210424'
        date_seems_trade_day = '20210217'
        date_too_early = '19890601'
        date_too_late = '20500105'
        date_christmas = '20201225'

        self.assertTrue(maybe_trade_day(date_trade))
        self.assertFalse(maybe_trade_day(date_holiday))
        self.assertFalse(maybe_trade_day(date_weekend))
        self.assertTrue(maybe_trade_day(date_seems_trade_day))
        self.assertTrue(maybe_trade_day(date_too_early))
        self.assertTrue(maybe_trade_day(date_too_late))
        self.assertTrue(maybe_trade_day(date_christmas))
        self.assertTrue(is_market_trade_day(date_trade))
        self.assertFalse(is_market_trade_day(date_holiday))
        self.assertFalse(is_market_trade_day(date_weekend))
        self.assertFalse(is_market_trade_day(date_seems_trade_day))
        self.assertTrue(is_market_trade_day(date_christmas))
        # self.assertFalse(is_market_trade_day(date_christmas, exchange='XHKG'))
        # test trade dates in some strange text formats:
        self.assertTrue(is_market_trade_day('2021-04-01'))
        self.assertTrue(is_market_trade_day(pd.to_datetime('2021-04-01 00:00:00')))
        self.assertTrue(is_market_trade_day('2021-04-01 00:00:00'))
        self.assertTrue(is_market_trade_day(datetime.date(2021, 4, 1)))
        self.assertTrue(is_market_trade_day(datetime.date(2021, 4, 1), 'SSE'))

        # raises when date out of range
        self.assertRaises(KeyError, is_market_trade_day, date_too_early)
        self.assertRaises(KeyError, is_market_trade_day, date_too_late)

        date_trade = pd.to_datetime('20210401')
        date_holiday = pd.to_datetime('20210102')
        date_weekend = pd.to_datetime('20210424')

        self.assertTrue(maybe_trade_day(date_trade))
        self.assertFalse(maybe_trade_day(date_holiday))
        self.assertFalse(maybe_trade_day(date_weekend))

    def test_weekday_name(self):
        """ test util func weekday_name()"""
        self.assertEqual(weekday_name(0), 'Monday')
        self.assertEqual(weekday_name(1), 'Tuesday')
        self.assertEqual(weekday_name(2), 'Wednesday')
        self.assertEqual(weekday_name(3), 'Thursday')
        self.assertEqual(weekday_name(4), 'Friday')
        self.assertEqual(weekday_name(5), 'Saturday')
        self.assertEqual(weekday_name(6), 'Sunday')

    def test_list_truncate(self):
        """ test util func list_truncate()"""
        the_list = [1, 2, 3, 4, 5]
        ls = list_truncate(the_list, 2, as_list=True)
        self.assertEqual(ls[0], [1, 2])
        self.assertEqual(ls[1], [3, 4])
        self.assertEqual(ls[2], [5])

        self.assertRaises(ValueError, list_truncate, the_list, 0)
        self.assertRaises(TypeError, list_truncate, 12, 0)
        self.assertRaises(TypeError, list_truncate, 0, the_list)

        ls = list_truncate(the_list, 7, as_list=True)
        self.assertEqual(ls[0], [1, 2, 3, 4, 5])

        g = list_truncate(the_list, 2, as_list=False)
        self.assertEqual(next(g), [1, 2])
        self.assertEqual(next(g), [3, 4])
        self.assertEqual(next(g), [5])
        self.assertRaises(StopIteration, next, g)

        g = list_truncate(the_list, 7, as_list=False)
        self.assertEqual(next(g), [1, 2, 3, 4, 5])

        # test if the function works with generator, NotImplemented
        # lst = (i for i in range(10))
        #
        # g = list_truncate(lst, 3, as_list=False)
        # self.assertEqual(next(g), [0, 1, 2])
        # self.assertEqual(next(g), [3, 4, 5])
        # self.assertEqual(next(g), [6, 7, 8])
        # self.assertEqual(next(g), [9])

    def test_maybe_trade_day(self):
        """ test util function maybe_trade_day()"""
        self.assertTrue(maybe_trade_day('20220104'))
        self.assertTrue(maybe_trade_day('2021-12-31'))
        self.assertTrue(maybe_trade_day(pd.to_datetime('2020/03/06')))

        self.assertFalse(maybe_trade_day('2020-01-01'))
        self.assertFalse(maybe_trade_day('2020/10/06'))

        self.assertRaises(Exception, maybe_trade_day, 'aaa')

    def test_prev_trade_day(self):
        """test the function prev_trade_day()
        """
        date_trade = '20210401'
        date_holiday = '20210102'
        prev_holiday = pd.to_datetime(date_holiday) - pd.Timedelta(2, 'd')
        date_weekend = '20210424'
        prev_weekend = pd.to_datetime(date_weekend) - pd.Timedelta(1, 'd')
        date_seems_trade_day = '20210217'
        prev_seems_trade_day = '20210217'
        date_too_early = '19890601'
        date_too_late = '20230105'
        date_christmas = '20201225'
        self.assertEqual(pd.to_datetime(prev_trade_day(date_trade)),
                         pd.to_datetime(date_trade))
        self.assertEqual(pd.to_datetime(prev_trade_day(date_holiday)),
                         pd.to_datetime(prev_holiday))
        self.assertEqual(pd.to_datetime(prev_trade_day(date_weekend)),
                         pd.to_datetime(prev_weekend))
        self.assertEqual(pd.to_datetime(prev_trade_day(date_seems_trade_day)),
                         pd.to_datetime(prev_seems_trade_day))
        self.assertEqual(pd.to_datetime(prev_trade_day(date_too_early)),
                         pd.to_datetime(date_too_early))
        self.assertEqual(pd.to_datetime(prev_trade_day(date_too_late)),
                         pd.to_datetime(date_too_late))
        self.assertEqual(pd.to_datetime(prev_trade_day(date_christmas)),
                         pd.to_datetime(date_christmas))

    def test_next_trade_day(self):
        """ test the function next_trade_day()
        """
        date_trade = '20210401'
        date_holiday = '20210102'
        next_holiday = pd.to_datetime(date_holiday) + pd.Timedelta(2, 'd')
        date_weekend = '20210424'
        next_weekend = pd.to_datetime(date_weekend) + pd.Timedelta(2, 'd')
        date_seems_trade_day = '20210217'
        next_seems_trade_day = '20210217'
        date_too_early = '19890601'
        date_too_late = '20230105'
        date_christmas = '20201225'
        self.assertEqual(pd.to_datetime(next_trade_day(date_trade)),
                         pd.to_datetime(date_trade))
        self.assertEqual(pd.to_datetime(next_trade_day(date_holiday)),
                         pd.to_datetime(next_holiday))
        self.assertEqual(pd.to_datetime(next_trade_day(date_weekend)),
                         pd.to_datetime(next_weekend))
        self.assertEqual(pd.to_datetime(next_trade_day(date_seems_trade_day)),
                         pd.to_datetime(next_seems_trade_day))
        self.assertEqual(pd.to_datetime(next_trade_day(date_too_early)),
                         pd.to_datetime(date_too_early))
        self.assertEqual(pd.to_datetime(next_trade_day(date_too_late)),
                         pd.to_datetime(date_too_late))
        self.assertEqual(pd.to_datetime(next_trade_day(date_christmas)),
                         pd.to_datetime(date_christmas))

    def test_nearest_market_trade_day(self):
        """ test the function nearest_market_trade_day()
        """
        date_trade = '20210401'
        date_holiday = '20210102'
        prev_holiday = pd.to_datetime(date_holiday) - pd.Timedelta(2, 'd')
        date_weekend = '20210424'
        prev_weekend = pd.to_datetime(date_weekend) - pd.Timedelta(1, 'd')
        date_seems_trade_day = '20210217'
        prev_seems_trade_day = pd.to_datetime(date_seems_trade_day) - pd.Timedelta(7, 'd')
        date_too_early = '19890601'
        date_too_late = '20990105'
        date_christmas = '20201225'
        prev_christmas_xhkg = '20201224'
        self.assertEqual(pd.to_datetime(nearest_market_trade_day(date_trade)),
                         pd.to_datetime(date_trade))
        self.assertEqual(pd.to_datetime(nearest_market_trade_day(date_holiday)),
                         pd.to_datetime(prev_holiday))
        self.assertEqual(pd.to_datetime(nearest_market_trade_day(date_weekend)),
                         pd.to_datetime(prev_weekend))
        self.assertEqual(pd.to_datetime(nearest_market_trade_day(date_seems_trade_day)),
                         pd.to_datetime(prev_seems_trade_day))
        self.assertEqual(pd.to_datetime(nearest_market_trade_day(date_too_early)),
                         None)
        self.assertEqual(pd.to_datetime(nearest_market_trade_day(date_too_late)),
                         None)
        self.assertEqual(pd.to_datetime(nearest_market_trade_day(date_christmas, 'SSE')),
                         pd.to_datetime(date_christmas))
        # self.assertEqual(pd.to_datetime(nearest_market_trade_day(date_christmas, 'XHKG')),
        #                  pd.to_datetime(prev_christmas_xhkg))

    def test_next_market_trade_day(self):
        """ test the function next_market_trade_day()
        """
        date_trade = '20210401'
        next_date = '20210402'
        date_holiday = '20210102'
        next_holiday = pd.to_datetime(date_holiday) + pd.Timedelta(2, 'd')
        date_weekend = '20210424'
        next_weekend = pd.to_datetime(date_weekend) + pd.Timedelta(2, 'd')
        date_seems_trade_day = '20210217'
        next_seems_trade_day = pd.to_datetime(date_seems_trade_day) + pd.Timedelta(1, 'd')
        date_too_early = '19890601'
        date_too_late = '20230105'
        date_christmas = '20201225'
        next_christmas_xhkg = '20201228'
        self.assertEqual(pd.to_datetime(next_market_trade_day(date_trade)),
                         pd.to_datetime(date_trade))
        self.assertEqual(pd.to_datetime(next_market_trade_day(date_trade, nearest_only=False)),
                         pd.to_datetime(next_date))
        self.assertEqual(pd.to_datetime(next_market_trade_day(date_holiday)),
                         pd.to_datetime(next_holiday))
        self.assertEqual(pd.to_datetime(next_market_trade_day(date_weekend)),
                         pd.to_datetime(next_weekend))
        self.assertEqual(pd.to_datetime(next_market_trade_day(date_seems_trade_day)),
                         pd.to_datetime(next_seems_trade_day))
        self.assertEqual(pd.to_datetime(next_market_trade_day(date_too_early)),
                         None)
        self.assertEqual(pd.to_datetime(next_market_trade_day(date_too_late)),
                         pd.to_datetime(date_too_late))  # data too late is not any more a problem
        self.assertEqual(pd.to_datetime(next_market_trade_day(date_christmas, 'SSE')),
                         pd.to_datetime(date_christmas))
        # self.assertEqual(pd.to_datetime(next_market_trade_day(date_christmas, 'XHKG')),
        #                  pd.to_datetime(next_christmas_xhkg))

    def test_is_number_like(self):
        """test the function: is_number_like()"""
        self.assertTrue(is_number_like(123))
        self.assertTrue(is_number_like(23.7))
        self.assertTrue(is_number_like('2.3'))
        self.assertTrue(is_number_like('-0.2'))
        self.assertTrue(is_number_like('0.03'))
        self.assertTrue(is_number_like('345'))
        self.assertTrue(is_number_like('-45.321'))

        self.assertFalse(is_number_like('-..023'))
        self.assertFalse(is_number_like('abc'))
        self.assertFalse(is_number_like('0.32a'))
        self.assertFalse(is_number_like('0-2'))

    def test_is_integer_like(self):
        """test the function: is_integer_like()"""
        self.assertTrue(is_integer_like(123))
        self.assertTrue(is_integer_like('123'))
        self.assertTrue(is_integer_like('-123'))
        self.assertTrue(is_integer_like('0'))
        self.assertTrue(is_integer_like('-0'))

        self.assertFalse(is_integer_like('0000'))
        self.assertFalse(is_integer_like(123.4))
        self.assertFalse(is_integer_like('123.4'))
        self.assertFalse(is_integer_like('-123.4'))
        self.assertFalse(is_integer_like('0.0'))
        self.assertFalse(is_integer_like('-0.0'))
        self.assertFalse(is_integer_like('0000.0'))
        self.assertFalse(is_integer_like('000.0'))
        self.assertFalse(is_integer_like('abc'))
        self.assertFalse(is_integer_like('0.32a'))
        self.assertFalse(is_integer_like('0-2'))

    def test_is_float_like(self):
        """test the function: is_float_like()"""
        self.assertTrue(is_float_like(123))
        self.assertTrue(is_float_like('123'))
        self.assertTrue(is_float_like('-123'))
        self.assertTrue(is_float_like('0'))
        self.assertTrue(is_float_like('-0'))
        self.assertTrue(is_float_like(123.4))
        self.assertTrue(is_float_like('123.4'))
        self.assertTrue(is_float_like('-123.4'))
        self.assertTrue(is_float_like('0.0'))
        self.assertTrue(is_float_like('-0.0'))

        self.assertFalse(is_float_like('0000.0'))
        self.assertFalse(is_float_like('000.0'))
        self.assertFalse(is_float_like('0000'))
        self.assertFalse(is_float_like('000'))
        self.assertFalse(is_float_like('abc'))
        self.assertFalse(is_float_like('0.32a'))
        self.assertFalse(is_float_like('0-2'))

    def test_is_stock_symbol_like(self):
        """ test functions is_cn_stock_symbol_like() and is_complete_cn_stock_symbol_like() """
        self.assertTrue(is_cn_stock_symbol_like('000001'))
        self.assertTrue(is_cn_stock_symbol_like('100001'))
        self.assertFalse(is_cn_stock_symbol_like('0000001'))
        self.assertFalse(is_cn_stock_symbol_like('00001'))
        self.assertFalse(is_cn_stock_symbol_like('00001a'))
        self.assertFalse(is_cn_stock_symbol_like('00001A'))
        self.assertFalse(is_cn_stock_symbol_like('00001.'))
        self.assertFalse(is_cn_stock_symbol_like('00001.0'))
        self.assertFalse(is_cn_stock_symbol_like('00001.00'))

        self.assertTrue(is_complete_cn_stock_symbol_like('000001.SZ'))
        self.assertTrue(is_complete_cn_stock_symbol_like('100001.SZ'))
        self.assertTrue(is_complete_cn_stock_symbol_like('000001.SH'))
        self.assertTrue(is_complete_cn_stock_symbol_like('100001.BJ'))
        self.assertFalse(is_complete_cn_stock_symbol_like('0000001.SZ'))
        self.assertFalse(is_complete_cn_stock_symbol_like('00001.SZ'))
        self.assertFalse(is_complete_cn_stock_symbol_like('000011'))
        self.assertFalse(is_complete_cn_stock_symbol_like('00001a.SZ'))
        self.assertFalse(is_complete_cn_stock_symbol_like('00001A.SZ'))
        self.assertFalse(is_complete_cn_stock_symbol_like('00001.ZH'))

    def test_retry_decorator(self):
        """ test the retry decorator"""
        print(f'test no retry needed functions')
        self.counter = 0

        @retry(RetryableError, tries=4, delay=0.1)
        def succeeds():
            self.counter += 1
            return 'success'

        r = succeeds()
        self.assertEqual(r, 'success')
        self.assertEqual(self.counter, 1)

        print(f'test retry only once')
        self.counter = 0

        @retry(RetryableError, tries=4, delay=0.1)
        def fails_once():
            self.counter += 1
            if self.counter < 2:
                raise RetryableError('failed')
            else:
                return 'success'

        r = fails_once()
        self.assertEqual(r, 'success')
        self.assertEqual(self.counter, 2)

        print(f'test retry limit is reached')
        self.counter = 0

        @retry(RetryableError, tries=4, delay=0.1)
        def always_fails():
            self.counter += 1
            raise RetryableError('failed')

        with self.assertRaises(RetryableError):
            always_fails()
        self.assertEqual(self.counter, 4)

        print(f'test checking for multiple exceptions')
        self.counter = 0

        @retry((RetryableError, AnotherRetryableError), tries=4, delay=0.1)
        def raise_multiple_exceptions():
            self.counter += 1
            if self.counter == 1:
                raise RetryableError('a retryable error')
            elif self.counter == 2:
                raise AnotherRetryableError('another retryable error')
            else:
                return 'success'

        r = raise_multiple_exceptions()
        self.assertEqual(r, 'success')
        self.assertEqual(self.counter, 3)

        print(f'does not retry when unexpected error occurs')

        @retry(RetryableError, tries=4, delay=0.1)
        def raise_unexpected_error():
            raise UnexpectedError('unexpected error')

        with self.assertRaises(UnexpectedError):
            raise_unexpected_error()

        print(f'test recording info with a logger')
        self.counter = 0
        # set up the logger object
        sh = logging.StreamHandler()
        logger = logging.getLogger(__name__)
        logger.addHandler(sh)

        @retry(RetryableError, tries=4, delay=0.1, logger=logger)
        def fails_once():
            self.counter += 1
            if self.counter < 2:
                raise RetryableError('failed')
            else:
                return 'success'

        fails_once()

    def test_lev_match(self):
        """ 测试字符串模糊匹配"""
        print(_lev_ratio('abc', 'ABC'))
        print(_lev_ratio('abcdefg', 'abDdefh'))
        print(_lev_ratio('中国移动', '中国移动总公司'))
        print(_lev_ratio('平安', '平安银行股份有限公司'))
        print(_lev_ratio('中国平安', '平安银行股份有限公司'))
        print(_lev_ratio('平安', '万科企业股份有限公司'))

        print(_partial_lev_ratio('abc', 'ABC'))
        print(_partial_lev_ratio('abcdefg', 'abDdefh'))
        print(_partial_lev_ratio('中国移动', '中国移动总公司'))
        print(_partial_lev_ratio('平安', '平安银行股份有限公司'))
        print(_partial_lev_ratio('中国平安', '平安银行股份有限公司'))
        print(_partial_lev_ratio('平安', '万科企业股份有限公司'))
        print(_partial_lev_ratio('常辅股份', '常州电站辅机股份有限公司'))

        print(_partial_lev_ratio('平?', '万科企业股份有限公司'))
        print(_partial_lev_ratio('常?股份', '常州电站辅机股份有限公司'))

    def test_wildcard_match(self):
        """ 测试字符串通配符匹配"""
        wordlist = ["color", "colour", "work", "working", "fox", "worker", "working"]
        print(_wildcard_match('col?r', wordlist))
        print(_wildcard_match('col*r', wordlist))
        print(_wildcard_match('wor.?', wordlist))
        print(_wildcard_match('?o?', wordlist))

    def test_match_ts_code(self):
        """ 测试匹配ts_code"""
        print(f"matching {'000001'}: \n{match_ts_code('000001')}")
        print(f"matching {'000001.CZC'}: \n{match_ts_code('000001.CZC')}")
        print(f"matching {'ABC.CZC'}: \n{match_ts_code('ABC.CZC')}")
        print(f"matching {'中国电信'}: \n{match_ts_code('中国电信')}")
        print(f"matching {'嘉实服务'}: \n{match_ts_code('嘉实服务')}")
        print(f"matching {'中?集团'}: \n{match_ts_code('中?集团')}")
        print(f"matching {'中*金'}: \n{match_ts_code('中*金')}")
        print(f"matching {'工商银行'}: \n{match_ts_code('工商银行')}")
        print(f"matching {'贵州钢绳'}: \n{match_ts_code('贵州钢绳')}")
        print(f"matching {'贵州钢绳'} with match_full_name: \n{match_ts_code('贵州钢绳', match_full_name=True)}")
        print(f"matching {'招商银行'} with asset_type = 'E, FD': \n{match_ts_code('招商银行', asset_types='E, FD')}")
        print(f"matching {'贵阳银行'} with asset_type = 'E, FT': \n{match_ts_code('贵阳银行', asset_types='E, FT')}")

    def test_rolling_window(self):
        """ 测试含税rolling_window()"""
        # test 1d array
        arr = np.array([1, 2, 3, 4, 5]).astype('int64')
        window = rolling_window(arr, window=3, axis=0)
        print(f'origin array: \n{arr}\n'
              f'rolling window: \n{window}')
        target = np.array([[1, 2, 3],
                           [2, 3, 4],
                           [3, 4, 5]])
        self.assertTrue(np.allclose(window, target))
        # test 1d array with int32 dtype
        arr = np.array([1, 2, 3, 4, 5]).astype('int32')
        window = rolling_window(arr, window=3, axis=0)
        print(f'origin array: \n{arr}\n'
              f'rolling window: \n{window}')
        target = np.array([[1, 2, 3],
                           [2, 3, 4],
                           [3, 4, 5]])
        self.assertTrue(np.allclose(window, target))
        # test 2d array
        arr = np.array([[1, 2, 3, 4],
                        [5, 6, 7, 8],
                        [9, 0, 1, 2]])
        window = rolling_window(arr, window=2, axis=0)
        print(f'origin array: \n{arr}\n'
              f'rolling window: \n{window}')
        target = np.array([[[1, 2, 3, 4],
                            [5, 6, 7, 8]],
                           [[5, 6, 7, 8],
                            [9, 0, 1, 2]]]).astype('int64')
        self.assertTrue(np.allclose(window, target))
        window = rolling_window(arr, window=3, axis=1)
        print(f'origin array: \n{arr}\n'
              f'rolling window: \n{window}')
        target = np.array([[[1, 2, 3],
                            [5, 6, 7],
                            [9, 0, 1]],
                           [[2, 3, 4],
                            [6, 7, 8],
                            [0, 1, 2]]])
        self.assertTrue(np.allclose(window, target))
        # test 2d array with int32 dtype
        arr = np.array([[1, 2, 3, 4],
                        [5, 6, 7, 8],
                        [9, 0, 1, 2]]).astype('int32')
        window = rolling_window(arr, window=2, axis=0)
        print(f'origin array: \n{arr}\n'
              f'rolling window: \n{window}')
        target = np.array([[[1, 2, 3, 4],
                            [5, 6, 7, 8]],
                           [[5, 6, 7, 8],
                            [9, 0, 1, 2]]])
        self.assertTrue(np.allclose(window, target))
        window = rolling_window(arr, window=3, axis=1)
        print(f'origin array: \n{arr}\n'
              f'rolling window: \n{window}')
        target = np.array([[[1, 2, 3],
                            [5, 6, 7],
                            [9, 0, 1]],
                           [[2, 3, 4],
                            [6, 7, 8],
                            [0, 1, 2]]])
        self.assertTrue(np.allclose(window, target))
        # test 3d array
        arr = np.array([[[1, 2, 3, 4],
                         [5, 6, 7, 8],
                         [9, 0, 1, 2]],

                        [[3, 4, 5, 6],
                         [7, 8, 9, 0],
                         [1, 2, 3, 4]],

                        [[5, 6, 7, 8],
                         [9, 0, 1, 2],
                         [3, 4, 5, 6]]])
        window = rolling_window(arr, window=2, axis=0)
        print(f'origin array: \n{arr}\n'
              f'rolling window: \n{window}')
        target = np.array([[[[1, 2, 3, 4],
                             [5, 6, 7, 8],
                             [9, 0, 1, 2]],

                            [[3, 4, 5, 6],
                             [7, 8, 9, 0],
                             [1, 2, 3, 4]]],

                           [[[3, 4, 5, 6],
                             [7, 8, 9, 0],
                             [1, 2, 3, 4]],

                            [[5, 6, 7, 8],
                             [9, 0, 1, 2],
                             [3, 4, 5, 6]]]])
        self.assertTrue(np.allclose(window, target))
        window = rolling_window(arr, window=2, axis=1)
        print(f'origin array: \n{arr}\n'
              f'rolling window: \n{window}')
        target = np.array([[[[1, 2, 3, 4],
                             [5, 6, 7, 8]],

                            [[3, 4, 5, 6],
                             [7, 8, 9, 0]],

                            [[5, 6, 7, 8],
                             [9, 0, 1, 2]]],

                           [[[5, 6, 7, 8],
                             [9, 0, 1, 2]],

                            [[7, 8, 9, 0],
                             [1, 2, 3, 4]],

                            [[9, 0, 1, 2],
                             [3, 4, 5, 6]]]])
        self.assertTrue(np.allclose(window, target))
        window = rolling_window(arr, window=3, axis=2)
        print(f'origin array: \n{arr}\n'
              f'rolling window: \n{window}')
        target = np.array([[[[1, 2, 3],
                             [5, 6, 7],
                             [9, 0, 1]],

                            [[3, 4, 5],
                             [7, 8, 9],
                             [1, 2, 3]],

                            [[5, 6, 7],
                             [9, 0, 1],
                             [3, 4, 5]]],

                           [[[2, 3, 4],
                             [6, 7, 8],
                             [0, 1, 2]],

                            [[4, 5, 6],
                             [8, 9, 0],
                             [2, 3, 4]],

                            [[6, 7, 8],
                             [0, 1, 2],
                             [4, 5, 6]]]
                           ])
        self.assertTrue(np.allclose(window, target))

        # test false input
        arr = np.array([[[1, 2, 3, 4],
                         [5, 6, 7, 8],
                         [9, 0, 1, 2]],

                        [[3, 4, 5, 6],
                         [7, 8, 9, 0],
                         [1, 2, 3, 4]],

                        [[5, 6, 7, 8],
                         [9, 0, 1, 2],
                         [3, 4, 5, 6]]])
        self.assertRaises(TypeError, rolling_window, 1, 1, 1)
        self.assertRaises(TypeError, rolling_window, 1, 's', 1)
        self.assertRaises(TypeError, rolling_window, 1, 1, 's')
        self.assertRaises(ValueError, rolling_window, arr, 1, -1)
        self.assertRaises(ValueError, rolling_window, arr, -1, -1)
        self.assertRaises(ValueError, rolling_window, arr, 5, 1)
        self.assertRaises(ValueError, rolling_window, arr, 2, 3)

    def test_reindent(self):
        """ 测试reindent函数"""
        string = 'a random string'
        mul_str = 'a multiple-line random string\n' \
                  ' a multiple-line random string\n' \
                  'a multiple-line random string'

        self.assertEqual(reindent(string, 4), '    a random string')
        self.assertEqual(reindent(mul_str, 4), '    a multiple-line random string\n'
                                               '    a multiple-line random string\n'
                                               '    a multiple-line random string')

        self.assertRaises(TypeError, reindent, 123, 123)
        self.assertRaises(TypeError, reindent, '123', '123')
        self.assertRaises(ValueError, reindent, '123', 123)

    def test_truncate_string(self):
        """ 测试函数truncates_string"""
        self.assertEqual(adjust_string_length('this is a long string', 23), 'this is a long string  ')
        self.assertEqual(adjust_string_length('this is a long string', 22), 'this is a long string ')
        self.assertEqual(adjust_string_length('this is a long string', 21), 'this is a long string')
        self.assertEqual(adjust_string_length('this is a long string', 20), 'this is a lo...tring')
        self.assertEqual(adjust_string_length('this is a long string', 19), 'this is a l...tring')
        self.assertEqual(adjust_string_length('this is a long string', 18), 'this is a ...tring')
        self.assertEqual(adjust_string_length('this is a long string', 17), 'this is a ...ring')
        self.assertEqual(adjust_string_length('this is a long string', 16), 'this is a...ring')
        self.assertEqual(adjust_string_length('this is a long string', 15), 'this is ...ring')
        self.assertEqual(adjust_string_length('this is a long string', 10), 'this ...ng')
        self.assertEqual(adjust_string_length('this is a long string', 7), 'thi...g')
        self.assertEqual(adjust_string_length('this is a long string', 6), 'th...g')
        self.assertEqual(adjust_string_length('this is a long string', 5), 't...g')
        self.assertEqual(adjust_string_length('this is a long string', 4), 't...')
        self.assertEqual(adjust_string_length('this is a long string', 3), 't..')
        self.assertEqual(adjust_string_length('this is a long string', 2), '..')
        self.assertEqual(adjust_string_length('this is a long string', 1), '.')

        self.assertEqual(adjust_string_length('this is a long string', 10, filler='*'), 'this ***ng')
        self.assertEqual(adjust_string_length('short string', 15), 'short string   ')
        self.assertEqual(adjust_string_length('中文字符string', 20, hans_aware=True),
                         '中文字符string      ')
        self.assertEqual(adjust_string_length('中文字符string', 20, hans_aware=False),
                         '中文字符string          ')
        self.assertEqual(adjust_string_length('中文字符string', 20, padder='*', hans_aware=True),
                         '中文字符string******')
        self.assertEqual(adjust_string_length('中国平安', 8, hans_aware=True),
                         '中国平安')
        self.assertEqual(adjust_string_length('中国龙A', 8, hans_aware=True),
                         '中国龙A ')
        self.assertEqual(adjust_string_length('ST中国', 8, hans_aware=True),
                         'ST中国  ')
        self.assertEqual(adjust_string_length('this is a 含有中文字符的 string', 20, hans_aware=False),
                         'this is a 含有...tring')
        self.assertEqual(adjust_string_length('this is a 含有中文字符的 string', 19, hans_aware=False),
                         'this is a 含...tring')
        self.assertEqual(adjust_string_length('this is a 含有中文字符的 string', 18, hans_aware=False),
                         'this is a ...tring')
        self.assertEqual(adjust_string_length('this is a 含有中文字符的 string', 17, hans_aware=False),
                         'this is a ...ring')
        self.assertEqual(adjust_string_length('this is a 含有中文字符的 string', 16, hans_aware=False),
                         'this is a...ring')
        self.assertEqual(adjust_string_length('this is a 含有中文字符的 string', 15, hans_aware=False),
                         'this is ...ring')
        self.assertEqual(adjust_string_length('this is a 含有中文字符的 string', 20, hans_aware=True),
                         'this is a 含...tring')
        self.assertEqual(adjust_string_length('this is a 含有中文字符的 string', 19, hans_aware=True),
                         'this is a 含..tring')
        self.assertEqual(adjust_string_length('this is a 含有中文字符的 string', 18, hans_aware=True),
                         'this is a ...tring')
        self.assertEqual(adjust_string_length('this is a 含有中文字符的 string', 17, hans_aware=True),
                         'this is a ...ring')
        self.assertEqual(adjust_string_length('this is a 含有中文字符的 string', 16, hans_aware=True),
                         'this is a...ring')
        self.assertEqual(adjust_string_length('this is a 含有中文字符的 string', 15, hans_aware=True),
                         'this is ...ring')
        self.assertEqual(adjust_string_length('含有中文字符的 string', 20, hans_aware=False),
                         '含有中文字符的 string      ')
        self.assertEqual(adjust_string_length('含有中文字符的 string', 10, hans_aware=False),
                         '含有中文字...ng')
        self.assertEqual(adjust_string_length('含有中文字符的 string', 9, hans_aware=False),
                         '含有中文...ng')
        self.assertEqual(adjust_string_length('含有中文字符的 string', 8, hans_aware=False),
                         '含有中文...g')
        self.assertEqual(adjust_string_length('含有中文字符的 string', 7, hans_aware=False),
                         '含有中...g')
        self.assertEqual(adjust_string_length('含有中文字符的 string', 6, hans_aware=False),
                         '含有...g')
        self.assertEqual(adjust_string_length('含有中文字符的 string', 5, hans_aware=False),
                         '含...g')
        self.assertEqual(adjust_string_length('含有中文字符的 string', 4, hans_aware=False),
                         '含...')
        self.assertEqual(adjust_string_length('含有中文字符的 string', 3, hans_aware=False),
                         '含..')
        self.assertEqual(adjust_string_length('含有中文字符的 string', 2, hans_aware=False),
                         '..')
        self.assertEqual(adjust_string_length('含有中文字符的 string', 1, hans_aware=False),
                         '.')
        self.assertEqual(adjust_string_length('含有中文字符的 string', 20, hans_aware=True),
                         '含有中文字符...tring')
        self.assertEqual(adjust_string_length('含有中文字符的 string', 10, hans_aware=True),
                         '含有中..ng')
        self.assertEqual(adjust_string_length('含有中文字符的 string', 9, hans_aware=True),
                         '含有...ng')
        self.assertEqual(adjust_string_length('含有中文字符的 string', 8, hans_aware=True),
                         '含有...g')
        self.assertEqual(adjust_string_length('含有中文字符的 string', 7, hans_aware=True),
                         '含有..g')
        self.assertEqual(adjust_string_length('含有中文字符的 string', 6, hans_aware=True),
                         '含...g')
        self.assertEqual(adjust_string_length('含有中文字符的 string', 5, hans_aware=True),
                         '含..g')
        self.assertEqual(adjust_string_length('含有中文字符的 string', 4, hans_aware=True),
                         '含..')
        self.assertEqual(adjust_string_length('含有中文字符的 string', 3, hans_aware=True),
                         '含.')
        self.assertEqual(adjust_string_length('含有中文字符的 string', 2, hans_aware=True),
                         '..')
        self.assertEqual(adjust_string_length('含有中文字符的 string', 1, hans_aware=True),
                         '.')
        self.assertEqual(adjust_string_length('中文字符占据2个位置', 14, hans_aware=True),
                         '中文字符..位置')
        self.assertEqual(adjust_string_length('中文字符占据2个位置', 13, hans_aware=True),
                         '中文字符.位置')
        self.assertEqual(
            adjust_string_length('[red]string[/red] [blue]with[/blue] [yellow]format[/yellow] [green]tags[/green]', 13,
                                 format_tags=True), '[red]string[/red] [blue]...[/blue][yellow][/yellow][green]ags['
                                                    '/green]')
        self.assertEqual(adjust_string_length(
            '[red]string[/red] [blue]with[/blue] [yellow]format[/yellow] [green]tags[/green]及[white]中文字符[/white]',
            13, hans_aware=False, format_tags=True),
                '[red]string[/red] [blue]...[/blue][yellow][/yellow][green][/green][white]文字符[/white]')
        self.assertEqual(adjust_string_length(
            '[red]string[/red] [blue]with[/blue] [yellow]format[/yellow] [green]tags[/green]及[white]中文字符[/white]',
            13, hans_aware=True, format_tags=True),
                '[red]string[/red] [blue]..[/blue][yellow][/yellow][green][/green][white]字符[/white]')
        self.assertEqual(adjust_string_length(
            '[red]完全[/red][blue]由[/blue][yellow]中文字符[/yellow][green]组成[/green]的[white]字符串[/white]', 10,
            hans_aware=False, format_tags=True),
                '[red]完全[/red][blue]由[/blue][yellow]中文..[/yellow][green].[/green][white]符串[/white]')
        self.assertEqual(adjust_string_length(
            '[red]完全[/red][blue]由[/blue][yellow]中文字符[/yellow][green]组成[/green]的[white]字符串[/white]', 13,
            hans_aware=True, format_tags=True),
                '[red]完全[/red][blue]由[/blue][yellow]中.[/yellow][green][/green][white]符串[/white]')
        self.assertEqual(adjust_string_length('this is a long string', 7, format_tags=True), 'thi...g')

        self.assertRaises(TypeError, adjust_string_length, 123, 10)
        self.assertRaises(TypeError, adjust_string_length, 123, 'this ia a string')
        self.assertRaises(ValueError, adjust_string_length, 'this is a long string', 5, filler='__')
        self.assertRaises(ValueError, adjust_string_length, 'this is a long string', 0)


if __name__ == '__main__':
    unittest.main()


# coding=utf-8
# ======================================
# File:     qt_code.py
# Author:   Jackie PENG
# Contact:  jackie.pengzhao@gmail.com
# Created:  2024-10-20
# Desc:
#   Definition of class QtCode, storing
# and managing different market symbols.
# ======================================


from typing import Optional, Union

from .utilfuncs import AVAILABLE_ASSET_TYPES


def _is_valid_code(code: str) -> bool:
    """ check if the code is legal

    """
    if not isinstance(code, str):
        return False
    # no space or special characters
    if any([c in code for c in ' ~!@#$%^&*()_+={}[]|\\:;\"\'<>,?/']):
        return False

    # max two parts separated by a dot and length of each part is at least 1
    parts = code.split('.')
    if len(parts) > 2:
        return False
    if len(parts) == 1 and len(parts[0]) < 4:
        return False

    if not all([len(p) > 0 for p in parts]):
        return False
    # if both parts are longer than 2, then return False
    if len(parts) == 2 and all([len(p) > 2 for p in parts]):
        return False

    return True


def _infer_symbol_market(code) -> Optional[tuple[str, str]]:
    """ infer the symbol and market from the code as a tuple:

    Parameters
    ----------
    code : str
        code or symbol

    Returns
    -------
    tuple(str, str), optional
        symbol and market
    """
    # if the code splits into only one part, then it is the symbol
    parts = code.split('.')
    if len(parts) == 1:
        symbol = parts[0]
        # infer market according to symbol
        if len(symbol) == 5 and all(c.isdigit() for c in symbol):
            # symbols like '00001' are HK stocks
            return symbol, 'HK'
        elif len(symbol) == 8 and all(c.isdigit() for c in symbol):
            # symbols like '00000001' are SH options
            return symbol, 'SH'
        elif len(symbol) == 6 and all(c.isdigit() for c in symbol):
            # symbols like '000001' are SH stocks or SZ stocks depending on the first digit
            if symbol[0] in ['6', '9']:
                return symbol, 'SH'
            elif symbol[0] in ['0', '3']:
                return symbol, 'SZ'
            else:
                return symbol, 'SH'
        elif len(symbol) < 6 and all(c.isdigit() for c in symbol[-4:]) and all(c.isalpha() for c in symbol[:-4]):
            # symbols like 'A0001' or 'CU2310' are futures
            initials = ''.join([c for c in symbol if c.isalpha()])
            if initials in ['CU', 'AL', 'AL-C/P', 'PB', 'ZN', 'ZN-C/P', 'SN', 'NI', 'SS', 'AU',
                            'AG', 'RB', 'HC', 'BU', 'RU', 'FU', 'SP', 'RU-C/P', 'WR']:
                return symbol, 'SHF'
            elif initials in ['A', 'B', 'M', 'M-C/P', 'Y', 'P', 'I', 'J', 'JM', 'C', 'CS', 'L', 'V', 'PP', 'EG',
                              'RR', 'EB', 'I-C/P', 'PG', 'JD', 'FB', 'BB', 'LH']:
                return symbol, 'DCE'
            elif initials in ['RM', 'OI', 'CF', 'TA', 'SR', 'MA', 'FG', 'ZC', 'CY', 'SA', 'PF', 'JR', 'RS', 'PM',
                              'WH', 'RI', 'LR', 'SF', 'SM', 'AP', 'CJ', 'UR', 'PK', 'PX', 'SH']:
                return symbol, 'ZCE'
            elif initials in ['IF', 'IH', 'IC', 'TF', 'T', 'TS']:
                return symbol, 'CFX'
            elif initials in ['SC', 'NR', 'LU', 'BC']:
                return symbol, 'INE'
            elif initials in ['SI', 'LC']:
                return symbol, 'GFE'
        elif len(symbol) <= 6 and all(c.isalpha() for c in symbol):
            # symbols like 'MSFT' or 'AAPL' are US stocks
            return symbol, 'US'
        else:
            # the rest are temporarily considered as SH stocks
            return parts[0], 'SH'
    else:
        # if both parts are pure letters, then the longer one is the symbol
        if all([p.isalpha() for p in parts]):
            return max(parts, key=len), min(parts, key=len)
        # if one part is pure letters and the other is pure digits, then the digit part is the symbol
        elif parts[0].isalpha() and parts[1].isdigit():
            return parts[1], parts[0]
        elif parts[1].isalpha() and parts[0].isdigit():
            return parts[0], parts[1]
        # if both parts are digits, then the longer one is the symbol
        elif parts[0].isdigit() and parts[1].isdigit():
            return max(parts, key=len), min(parts, key=len)
        # if both parts are mixed, then the longer one is the symbol
        else:
            return max(parts, key=len), min(parts, key=len)


def _infer_asset_type(symbol, market):
    """ infer the asset type according to symbol and market

    """
    # rules are different from market to market
    if market == 'SH':
        if len(symbol) > 6 and all(c.isdigit() for c in symbol):
            return 'OPT'
        elif symbol[0:3] in ['600', '601', '603', '900']:
            return 'E'
        elif symbol[0:3] in ['000']:
            return 'IDX'
        elif (symbol[0:3] > '100') and (symbol < '270'):
            return 'BOND'
        elif symbol[0] == '5':
            return 'FD'
        else:
            return 'E'
    elif market == 'SZ':
        if symbol[0:2] in ['00', '03', '07', '08', '09']:
            return 'E'
        elif symbol[0:2] in ['10', '11', '12', '13']:
            return 'BOND'
        elif symbol[0:2] in ['17', '18']:
            return 'FD'
        elif symbol[0:2] in ['39']:
            return 'IDX'
        else:
            return 'E'
    elif market == 'BJ':
        return 'E'
    elif market == 'OF':
        return 'FD'
    elif market in ['DCE', 'ZCE', 'CFX', 'INE', 'GFE', 'SHF']:
        return 'FT'

    elif market == 'HK':
        if symbol[0:1] in ['0', '1', '2', '3', '4', '5', '6', '7', '8', '9']:
            return 'E'
        else:
            return 'IDX'

    elif market == 'US':
        return 'E'

    else:
        return 'E'


class QtCode:
    """ special string class for market symbols, with properties
    that can be used to identify market types, and easy access of
    stock symbols and asset types

    """

    market_names = {
        'SZ': 'Shenzhen',
        'SH': 'Shanghai',
        'BJ': 'Beijing',
        'OF': 'Out market Fund',
        'DCE': 'Dalian Commodity Exchange',
        'ZCE': 'Zhengzhou Commodity Exchange',
        'CFX': 'China Financial Exchange',
        'INE': 'International Exchange',
        'GFE': 'Guangzhou Futures Exchange',
        'SHF': 'Shanghai Futures Exchange',
        'HK': 'Hong Kong',
        'US': 'United States',
    }

    def __init__(self, code_or_symbol=None, *, market=None, asset_type=None):
        """ initialize the QtCode object

        Parameters
        ----------
        code_or_symbol : str
            the market symbol or complete code with market separated by a dot
        market : str
            the market of the symbol
        asset_type : str
            the asset type of the symbol
        """
        self._symbol = None
        self._market = None
        self._asset_type = None
        self._code = None

        if _is_valid_code(code_or_symbol):
            if market is None:
                symbol, market = _infer_symbol_market(code_or_symbol)
            else:
                symbol, _ = _infer_symbol_market(code_or_symbol)

            if asset_type is None:
                asset_type = _infer_asset_type(symbol, market)

            self._parse_code(symbol, market, asset_type)
        else:
            raise ValueError(f'Invalid code: {code_or_symbol}')

    def _parse_code(self, symbol, market, asset_type) -> None:
        """ check if symbol matches market and asset type
        if so, return the code, otherwise raise ValueError
        """
        if asset_type not in AVAILABLE_ASSET_TYPES:
            raise ValueError(f'Invalid asset type: {asset_type}')

        if market not in self.market_names:
            raise ValueError(f'Invalid market: {market}')

        self._symbol = symbol
        self._market = market
        self._asset_type = asset_type
        self._code = f'{symbol}.{market}'

        return None

    @property
    def code(self):
        return self._code

    @property
    def market(self):
        return self._market

    @property
    def market_name(self):
        return self.market_names.get(self._market, None)

    @property
    def symbol(self):
        return self._symbol

    @property
    def asset_type(self):
        return self._asset_type

    def upper(self):
        return self.code.upper()

    def lower(self):
        return self.code.lower()

    def title(self):
        return self.code.title()

    def __get__(self, instance, owner):
        return self.code

    def __iter__(self):
        return iter(self.code)

    def __str__(self):
        return self.code

    def __repr__(self):
        return self.code

    def __eq__(self, other):
        return self.code == other

    def __hash__(self):
        return hash(self.code)

    def __ne__(self, other):
        return self.code != other

    def __lt__(self, other):
        return self.code < other

    def __le__(self, other):
        return self.code <= other

    def __gt__(self, other):
        return self.code > other

    def __ge__(self, other):
        return self.code >= other

    def __add__(self, other):
        return self.code + other

    def __radd__(self, other):
        return other + self.code

    def __mul__(self, other):
        return self.code * other

    def __rmul__(self, other):
        return other * self.code

    def __contains__(self, item):
        return item in self.code

    def __getitem__(self, item):
        return self.code[item]

    def __setitem__(self, key, value):
        self.code[key] = value

    def __delitem__(self, key):
        del self.code[key]

    def __len__(self):
        return len(self.code)

    def __iter__(self):
        return iter(self.code)

    def __reversed__(self):
        return reversed(self.code)

    def __contains__(self, item: object) -> object:
        return item in self.code

    def find(self, sub, start=None, end=None):
        return self.code.find(sub, start, end)

    def rfind(self, sub, start=None, end=None):
        return self.code.rfind(sub, start, end)

    def index(self, sub, start=None, end=None):
        return self.code.index(sub, start, end)

    def rindex(self, sub, start=None, end=None):
        return self.code.rindex(sub, start, end)

    def count(self, sub, start=None, end=None):
        return self.code.count(sub, start, end)



# coding=utf-8
# ======================================
# File:     test_hp_visual_spec.py
# Author:   Jackie PENG
# Contact:  jackie.pengzhao@gmail.com
# Created:  2025-03-13
# Desc:
#   Unittest for HistoryPanel 可视化规格：图表类型注册表与规格生成器（Phase 1）.
# ======================================

import unittest
import numpy as np
import pandas as pd

from qteasy.history import HistoryPanel
from qteasy.hp_visual_spec import (
    ChartTypeInfo,
    ChartTypeRegistry,
    get_chart_type_registry,
    build_kline_spec,
    build_volume_spec,
    build_macd_spec,
    build_line_spec,
)
from qteasy.hp_visual_render import (
    render_spec,
    render_kline_spec,
    render_volume_spec,
    render_macd_spec,
    render_line_spec,
)


def _make_hp(htypes: list, n_shares: int = 1, n_dates: int = 20) -> HistoryPanel:
    """构造小型 HistoryPanel，形状 (n_shares, n_dates, len(htypes))，数值为简单序列。"""
    n = len(htypes)
    values = np.tile(np.arange(n_dates, dtype=float).reshape(1, -1), (n_shares, 1))
    values = np.stack([values + i * 0.1 for i in range(n)], axis=2)
    shares = [f'S{i:03d}' for i in range(n_shares)]
    rows = pd.date_range('2020-01-01', periods=n_dates, freq='B')
    return HistoryPanel(values=values, levels=shares, rows=rows, columns=htypes)


class TestChartTypeRegistry(unittest.TestCase):
    """注册表：get_applicable_types 返回类型及重要性."""

    def setUp(self):
        self.registry = get_chart_type_registry()
        print('\n[TestChartTypeRegistry] using default registry')

    def test_registry_ohlc_returns_kline_main(self):
        print('\n[P1-1] 注册表：仅 OHLC 返回 kline(main)，无 volume/macd')
        htypes = ['open', 'high', 'low', 'close']
        result = self.registry.get_applicable_types(htypes)
        ids = [t.id for t in result]
        self.assertIn('kline', ids)
        k = next(t for t in result if t.id == 'kline')
        self.assertEqual(k.importance, 'main')
        self.assertNotIn('volume', ids)
        self.assertNotIn('macd', ids)
        print('  applicable:', ids, 'kline importance:', k.importance)

    def test_registry_ohlc_vol_returns_kline_and_volume(self):
        print('\n[P1-2] 注册表：OHLC+vol 返回 kline(main)、volume(secondary)，顺序 kline 在前')
        htypes = ['open', 'high', 'low', 'close', 'vol']
        result = self.registry.get_applicable_types(htypes)
        ids = [t.id for t in result]
        self.assertIn('kline', ids)
        self.assertIn('volume', ids)
        self.assertLess(ids.index('kline'), ids.index('volume'))
        k = next(t for t in result if t.id == 'kline')
        v = next(t for t in result if t.id == 'volume')
        self.assertEqual(k.importance, 'main')
        self.assertEqual(v.importance, 'secondary')
        print('  order:', ids, 'kline=', k.importance, 'volume=', v.importance)

    def test_registry_only_open_high_returns_line_not_kline(self):
        print('\n[P1-3] 注册表：仅 open+high 不返回 kline，折线适用 open/high，importance main')
        htypes = ['open', 'high']
        result = self.registry.get_applicable_types(htypes)
        ids = [t.id for t in result]
        self.assertNotIn('kline', ids)
        self.assertIn('line', ids)
        line = next(t for t in result if t.id == 'line')
        self.assertEqual(line.importance, 'main')
        print('  applicable:', ids, 'line importance:', line.importance)

    def test_registry_only_close_returns_line_main(self):
        print('\n[P1-4] 注册表：仅 close 返回 line(main)，无 kline/volume/macd')
        htypes = ['close']
        result = self.registry.get_applicable_types(htypes)
        ids = [t.id for t in result]
        self.assertNotIn('kline', ids)
        self.assertNotIn('volume', ids)
        self.assertNotIn('macd', ids)
        self.assertIn('line', ids)
        line = next(t for t in result if t.id == 'line')
        self.assertEqual(line.importance, 'main')
        print('  applicable:', ids)

    def test_registry_ohlc_vol_macd_returns_three_types(self):
        print('\n[P1-5] 注册表：OHLC+vol+MACD 模式匹配返回 kline(main)、volume、macd(secondary)')
        htypes = [
            'open', 'high', 'low', 'close', 'vol',
            'macd_12_26_9', 'macd_signal_12_26_9', 'macd_hist_12_26_9',
        ]
        result = self.registry.get_applicable_types(htypes)
        ids = [t.id for t in result]
        self.assertIn('kline', ids)
        self.assertIn('volume', ids)
        self.assertIn('macd', ids)
        self.assertLess(ids.index('kline'), ids.index('volume'))
        self.assertLess(ids.index('volume'), ids.index('macd'))
        print('  applicable:', ids)


class TestKlineSpecBuilder(unittest.TestCase):
    """K 线规格生成器：OHLC 必选，不含 vol；仅 open+high 或缺 close 返回 None."""

    def test_kline_spec_ohlc_no_vol(self):
        print('\n[P1-6] K 线规格：OHLC 无 vol，片段不含成交量，chart_type=kline importance=main')
        hp = _make_hp(['open', 'high', 'low', 'close'])
        spec = build_kline_spec(hp)
        self.assertIsNotNone(spec)
        self.assertEqual(spec['chart_type'], 'kline')
        self.assertEqual(spec['importance'], 'main')
        self.assertNotIn('vol', spec.get('data', {}))
        self.assertNotIn('volume', spec.get('data', {}))
        self.assertIn('open', spec['data'])
        self.assertIn('close', spec['data'])
        print('  chart_type:', spec['chart_type'], 'keys:', list(spec['data'].keys()))

    def test_kline_spec_ohlc_vol_separate_volume_spec(self):
        print('\n[P1-7] K 线+成交量：K 线片段仅 OHLC，volume 由 build_volume_spec 单独返回')
        hp = _make_hp(['open', 'high', 'low', 'close', 'vol'])
        kspec = build_kline_spec(hp)
        vspec = build_volume_spec(hp)
        self.assertIsNotNone(kspec)
        self.assertIsNotNone(vspec)
        self.assertNotIn('vol', kspec['data'])
        self.assertNotIn('volume', kspec['data'])
        self.assertIn('vol', vspec['data'])
        print('  kline data keys:', list(kspec['data'].keys()), 'volume data keys:', list(vspec['data'].keys()))

    def test_kline_spec_only_open_high_returns_none(self):
        print('\n[P1-8] K 线规格：仅 open+high 不产出 K 线，返回 None')
        hp = _make_hp(['open', 'high'])
        spec = build_kline_spec(hp)
        self.assertIsNone(spec)
        line_spec = build_line_spec(hp)
        self.assertIsNotNone(line_spec)
        self.assertIn('open', line_spec['htypes_used'])
        self.assertIn('high', line_spec['htypes_used'])
        print('  kline_spec:', spec, 'line_spec htypes_used:', line_spec['htypes_used'])

    def test_kline_spec_missing_close_returns_none(self):
        print('\n[P1-9] K 线规格：缺 close 不产出 K 线')
        hp = _make_hp(['open', 'high', 'low'])
        spec = build_kline_spec(hp)
        self.assertIsNone(spec)
        print('  build_kline_spec(hp):', spec)


class TestVolumeSpecBuilder(unittest.TestCase):
    """成交量规格生成器."""

    def test_volume_spec_has_vol(self):
        print('\n[P1-10] 成交量规格：HP 含 vol 则返回 volume 片段，importance=secondary')
        hp = _make_hp(['close', 'vol'])
        spec = build_volume_spec(hp)
        self.assertIsNotNone(spec)
        self.assertEqual(spec['chart_type'], 'volume')
        self.assertEqual(spec['importance'], 'secondary')
        print('  chart_type:', spec['chart_type'], 'importance:', spec['importance'])

    def test_volume_spec_no_vol_returns_none(self):
        print('\n[P1-11] 成交量规格：HP 不含 vol/volume 返回 None')
        hp = _make_hp(['open', 'high', 'low', 'close'])
        spec = build_volume_spec(hp)
        self.assertIsNone(spec)
        print('  build_volume_spec(hp):', spec)


class TestMacdSpecBuilder(unittest.TestCase):
    """MACD 规格生成器：模式匹配三列，缺一列则 None."""

    def test_macd_spec_three_cols_ok(self):
        print('\n[P1-12] MACD 规格：与 macd() 输出一致的三列齐全则非 None，能区分柱/线')
        htypes = ['macd_12_26_9', 'macd_signal_12_26_9', 'macd_hist_12_26_9']
        hp = _make_hp(htypes)
        spec = build_macd_spec(hp)
        self.assertIsNotNone(spec)
        self.assertEqual(spec['chart_type'], 'macd')
        self.assertIn('series_kind', spec)
        for col in htypes:
            self.assertIn(col, spec['data'])
            self.assertIn(spec['series_kind'][col], ('line', 'bar'))
        print('  series_kind:', spec['series_kind'])

    def test_macd_spec_missing_one_returns_none(self):
        print('\n[P1-13] MACD 规格：仅两列则 None')
        hp = _make_hp(['macd_12_26_9', 'macd_signal_12_26_9'])
        spec = build_macd_spec(hp)
        self.assertIsNone(spec)
        print('  build_macd_spec(hp):', spec)


class TestLineSpecBuilder(unittest.TestCase):
    """折线规格：仅 close、open+high 等."""

    def test_line_spec_only_close(self):
        print('\n[P1-14] 折线规格：仅 close 可产出折线片段，无 K 线时 importance=main')
        hp = _make_hp(['close'])
        spec = build_line_spec(hp)
        self.assertIsNotNone(spec)
        self.assertIn('close', spec['htypes_used'])
        self.assertEqual(spec['importance'], 'main')
        print('  htypes_used:', spec['htypes_used'], 'importance:', spec['importance'])

    def test_line_spec_open_high_no_ohlc(self):
        print('\n[P1-15] 折线规格：仅 open+high 时折线覆盖两条线，K 线规格为 None')
        hp = _make_hp(['open', 'high'])
        kspec = build_kline_spec(hp)
        line_spec = build_line_spec(hp)
        self.assertIsNone(kspec)
        self.assertIsNotNone(line_spec)
        self.assertIn('open', line_spec['htypes_used'])
        self.assertIn('high', line_spec['htypes_used'])
        self.assertEqual(len(line_spec['data']), 2)
        print('  line_spec htypes_used:', line_spec['htypes_used'])


class TestSpecDataFromHpOnly(unittest.TestCase):
    """规格数据仅来自 HP 切片，无额外计算."""

    def test_spec_values_match_hp_values(self):
        print('\n[P1-16] 规格数据与 hp.values 对应切片一致')
        hp = _make_hp(['open', 'high', 'low', 'close'], n_shares=2, n_dates=10)
        kspec = build_kline_spec(hp)
        self.assertIsNotNone(kspec)
        for h in ['open', 'close']:
            col_idx = hp.htypes.index(h)
            expected = hp.values[:, :, col_idx].astype(float)
            np.testing.assert_array_almost_equal(kspec['data'][h], expected)
        print('  kline data open shape:', kspec['data']['open'].shape, 'hp.values slice shape:', hp.values[:, :, 0].shape)


class TestKlineOptionalMaBbands(unittest.TestCase):
    """K 线可选 ma/bbands 叠加；仅 ma 或仅 bbands 无 OHLC 不产出 K 线."""

    def test_kline_spec_ohlc_with_ma_or_bbands(self):
        print('\n[P1-17] K 线规格：OHLC+ma/bbands 可选叠加，htypes_used 含对应列名')
        hp = _make_hp(['open', 'high', 'low', 'close', 'sma_20'])
        spec = build_kline_spec(hp)
        self.assertIsNotNone(spec)
        self.assertIn('sma_20', spec['htypes_used'])
        self.assertIn('sma_20', spec['data'])
        hp2 = _make_hp(['open', 'high', 'low', 'close', 'bbands_upper_20_2_2', 'bbands_middle_20_2_2', 'bbands_lower_20_2_2'])
        spec2 = build_kline_spec(hp2)
        self.assertIsNotNone(spec2)
        for b in ['bbands_upper_20_2_2', 'bbands_middle_20_2_2', 'bbands_lower_20_2_2']:
            self.assertIn(b, spec2['htypes_used'])
        print('  kline with sma_20 htypes_used:', spec['htypes_used'])

    def test_kline_spec_only_ma_or_only_bbands_returns_none(self):
        print('\n[P1-18] 仅 ma 或仅 bbands 无 OHLC 不产出 K 线，折线可承接')
        hp_ma = _make_hp(['sma_20'])
        spec_ma = build_kline_spec(hp_ma)
        self.assertIsNone(spec_ma)
        line_ma = build_line_spec(hp_ma)
        self.assertIsNotNone(line_ma)
        self.assertIn('sma_20', line_ma['htypes_used'])
        hp_bb = _make_hp(['bbands_upper_20_2_2', 'bbands_middle_20_2_2', 'bbands_lower_20_2_2'])
        spec_bb = build_kline_spec(hp_bb)
        self.assertIsNone(spec_bb)
        line_bb = build_line_spec(hp_bb)
        self.assertIsNotNone(line_bb)
        self.assertEqual(len(line_bb['htypes_used']), 3)
        print('  only sma_20 -> kline:', spec_ma, 'line htypes:', line_ma['htypes_used'])


class TestStaticRender(unittest.TestCase):
    """Phase 2：静态渲染层 — 给定规格+数据→Figure，返回类型与轴数量符合预期。"""

    def test_render_kline_returns_figure_one_axes(self):
        print('\n[Phase2] 单类型单 share：K 线规格渲染返回 Figure，至少 1 个 axes')
        hp = _make_hp(['open', 'high', 'low', 'close'])
        spec = build_kline_spec(hp)
        self.assertIsNotNone(spec)
        fig = render_kline_spec(spec)
        self.assertIsNotNone(fig)
        self.assertGreaterEqual(len(fig.axes), 1)
        print('  figure type:', type(fig).__name__, 'n_axes:', len(fig.axes))

    def test_render_volume_returns_figure_one_axes(self):
        print('\n[Phase2] 成交量规格渲染返回 Figure，至少 1 个 axes')
        hp = _make_hp(['close', 'vol'])
        spec = build_volume_spec(hp)
        self.assertIsNotNone(spec)
        fig = render_volume_spec(spec)
        self.assertIsNotNone(fig)
        self.assertGreaterEqual(len(fig.axes), 1)
        print('  n_axes:', len(fig.axes))

    def test_render_macd_returns_figure_one_axes(self):
        print('\n[Phase2] MACD 规格渲染返回 Figure，至少 1 个 axes')
        hp = _make_hp(['macd_12_26_9', 'macd_signal_12_26_9', 'macd_hist_12_26_9'])
        spec = build_macd_spec(hp)
        self.assertIsNotNone(spec)
        fig = render_macd_spec(spec)
        self.assertIsNotNone(fig)
        self.assertGreaterEqual(len(fig.axes), 1)
        print('  n_axes:', len(fig.axes))

    def test_render_line_returns_figure_one_axes(self):
        print('\n[Phase2] 折线规格渲染返回 Figure，至少 1 个 axes')
        hp = _make_hp(['close'])
        spec = build_line_spec(hp)
        self.assertIsNotNone(spec)
        fig = render_line_spec(spec)
        self.assertIsNotNone(fig)
        self.assertGreaterEqual(len(fig.axes), 1)
        print('  n_axes:', len(fig.axes))

    def test_render_spec_dispatches_by_chart_type(self):
        print('\n[Phase2] render_spec 按 chart_type 分发，返回 Figure')
        hp = _make_hp(['open', 'high', 'low', 'close'])
        spec = build_kline_spec(hp)
        fig = render_spec(spec)
        self.assertIsNotNone(fig)
        self.assertGreaterEqual(len(fig.axes), 1)
        hp2 = _make_hp(['close'])
        spec2 = build_line_spec(hp2)
        fig2 = render_spec(spec2)
        self.assertIsNotNone(fig2)
        print('  kline fig axes:', len(fig.axes), 'line fig axes:', len(fig2.axes))


class TestPlotOrchestrator(unittest.TestCase):
    """Phase 3：hp.plot() 编排器 — 1/2/多 shares、分组与组内子图数量."""

    def test_plot_one_share_returns_figure(self):
        print('\n[Phase3] 1 share：hp.plot() 返回 Figure，组内子图数量=适用图表类型数')
        hp = _make_hp(['open', 'high', 'low', 'close'], n_shares=1)
        fig = hp.plot()
        self.assertIsNotNone(fig)
        self.assertGreaterEqual(len(fig.axes), 1)
        print('  n_axes:', len(fig.axes))

    def test_plot_only_close_returns_one_subplot(self):
        print('\n[Phase3] 仅 close：仅 line 类型，1 个子图')
        hp = _make_hp(['close'], n_shares=1)
        fig = hp.plot()
        self.assertIsNotNone(fig)
        self.assertEqual(len(fig.axes), 1)
        print('  n_axes:', len(fig.axes))

    def test_plot_ohlc_vol_two_subplots(self):
        print('\n[Phase3] OHLC+vol：kline+volume，2 个子图')
        hp = _make_hp(['open', 'high', 'low', 'close', 'vol'], n_shares=1)
        fig = hp.plot()
        self.assertIsNotNone(fig)
        self.assertGreaterEqual(len(fig.axes), 2)
        print('  n_axes:', len(fig.axes))

    def test_plot_two_shares_stack_two_groups(self):
        print('\n[Phase3] 2 shares layout=stack：2 组，每组子图数=类型数')
        hp = _make_hp(['open', 'high', 'low', 'close'], n_shares=2)
        fig = hp.plot(layout='stack')
        self.assertIsNotNone(fig)
        self.assertGreaterEqual(len(fig.axes), 1)
        print('  n_axes:', len(fig.axes))


class TestHighlight(unittest.TestCase):
    """Phase 4：highlight 子模块 — condition + style，接入 hp.plot(..., highlight=...). 对应计划 8.7."""

    def test_p4_1_highlight_none_no_error(self):
        print('\n[P4-1] highlight=None 不报错，正常出图')
        hp = _make_hp(['close'], n_shares=1)
        fig = hp.plot(highlight=None)
        self.assertIsNotNone(fig)
        self.assertEqual(len(fig.axes), 1)
        ax = fig.axes[0]
        self.assertEqual(len(ax.collections), 0)
        print('  n_axes:', len(fig.axes), 'n_collections:', len(ax.collections))

    def test_p4_2_highlight_max_string_line(self):
        print('\n[P4-2] highlight=\"max\" 折线图，仅 close 最大值点高亮')
        hp = _make_hp(['close'], n_shares=1)
        fig = hp.plot(highlight='max')
        self.assertIsNotNone(fig)
        ax = fig.axes[0]
        collections = [c for c in ax.collections if hasattr(c, 'get_sizes')]
        self.assertGreaterEqual(len(collections), 1)
        print('  n_collections:', len(collections))

    def test_p4_3_highlight_condition_min_line(self):
        print('\n[P4-3] highlight={\"condition\": \"min\"} 折线图，最小值点高亮')
        hp = _make_hp(['close'], n_shares=1)
        spec = build_line_spec(hp)
        self.assertIsNotNone(spec)
        spec = dict(spec)
        spec['highlight'] = {'condition': 'min'}
        fig = render_line_spec(spec)
        self.assertIsNotNone(fig)
        ax = fig.axes[0]
        self.assertGreaterEqual(len(ax.collections), 1)
        print('  n_collections:', len(ax.collections))

    def test_p4_4_highlight_with_style(self):
        print('\n[P4-4] highlight 含 style，高亮点颜色与大小与 style 一致')
        hp = _make_hp(['close'], n_shares=1, n_dates=10)
        spec = build_line_spec(hp)
        spec = dict(spec)
        spec['highlight'] = {'condition': 'max', 'style': {'color': 'red', 's': 60}}
        fig = render_line_spec(spec)
        self.assertIsNotNone(fig)
        ax = fig.axes[0]
        self.assertGreaterEqual(len(ax.collections), 1)
        coll = ax.collections[0]
        self.assertEqual(coll.get_facecolor().shape[0], 1)
        self.assertAlmostEqual(coll.get_sizes()[0], 60)
        print('  collection sizes:', coll.get_sizes(), 'facecolor:', coll.get_facecolor())

    def test_p4_5_highlight_condition_bool_array(self):
        print('\n[P4-5] condition 为布尔数组，仅 True 处高亮')
        hp = _make_hp(['close'], n_shares=1, n_dates=15)
        spec = build_line_spec(hp)
        self.assertIsNotNone(spec)
        spec = dict(spec)
        mask = np.zeros(15, dtype=bool)
        mask[7] = True
        spec['highlight'] = {'condition': mask}
        fig = render_line_spec(spec)
        self.assertIsNotNone(fig)
        ax = fig.axes[0]
        self.assertGreaterEqual(len(ax.collections), 1)
        offs = ax.collections[0].get_offsets()
        self.assertEqual(offs.shape[0], 1)
        self.assertEqual(int(offs[0, 0]), 7)
        print('  single scatter at index:', int(offs[0, 0]))

    def test_p4_6_multi_group_stack_highlight(self):
        print('\n[P4-6] 多组 stack 时 highlight=\"max\"，每组折线各自 close 最大值高亮')
        hp = _make_hp(['close'], n_shares=2, n_dates=20)
        fig = hp.plot(layout='stack', highlight='max')
        self.assertIsNotNone(fig)
        self.assertEqual(len(fig.axes), 2)
        for i, ax in enumerate(fig.axes):
            collections = [c for c in ax.collections if hasattr(c, 'get_sizes')]
            self.assertGreaterEqual(len(collections), 1, f'group {i} should have highlight scatter')
        print('  n_axes:', len(fig.axes), 'each has scatter')

    def test_plot_with_highlight_max_dict(self):
        print('\n[Phase4] hp.plot(..., highlight={\"condition\": \"max\"}) 不报错，返回 Figure')
        hp = _make_hp(['close'], n_shares=1)
        fig = hp.plot(highlight={'condition': 'max'})
        self.assertIsNotNone(fig)
        self.assertEqual(len(fig.axes), 1)
        ax = fig.axes[0]
        collections = [c for c in ax.collections if hasattr(c, 'get_sizes')]
        self.assertGreaterEqual(len(collections), 1)
        print('  n_axes:', len(fig.axes), 'n_collections:', len(ax.collections))

    def test_render_line_spec_with_highlight_min(self):
        print('\n[Phase4] render_line_spec(spec with highlight condition=min) 绘制高亮点')
        hp = _make_hp(['close'], n_shares=1)
        spec = build_line_spec(hp)
        self.assertIsNotNone(spec)
        spec = dict(spec)
        spec['highlight'] = {'condition': 'min', 'style': {'color': 'red', 'marker': 'o'}}
        fig = render_line_spec(spec)
        self.assertIsNotNone(fig)
        ax = fig.axes[0]
        self.assertGreaterEqual(len(ax.collections), 1)
        print('  n_collections:', len(ax.collections))


class TestInteractiveBackend(unittest.TestCase):
    """Phase 5：hp.plot(..., interactive=True) 返回 Plotly Figure，结构可用."""

    def test_plot_interactive_true_returns_figure(self):
        print('\n[Phase5] hp.plot(interactive=True) 返回 Figure，结构可用')
        hp = _make_hp(['close'], n_shares=1)
        try:
            fig = hp.plot(interactive=True)
        except RuntimeError as e:
            if 'plotly' in str(e).lower():
                self.skipTest('plotly not installed, skip interactive test')
            raise
        self.assertIsNotNone(fig)
        if hasattr(fig, 'data'):
            self.assertGreaterEqual(len(fig.data), 1, 'Plotly figure should have at least one trace')
            print('  plotly fig.data traces:', len(fig.data))
        else:
            self.assertGreaterEqual(len(fig.axes), 1)
            print('  n_axes:', len(fig.axes))

    def test_plot_interactive_two_shares_stack(self):
        print('\n[Phase5] 2 shares layout=stack interactive 返回多 trace、多组')
        hp = _make_hp(['close'], n_shares=2)
        try:
            fig = hp.plot(interactive=True, layout='stack')
        except RuntimeError as e:
            if 'plotly' in str(e).lower():
                self.skipTest('plotly not installed')
            raise
        self.assertIsNotNone(fig)
        if hasattr(fig, 'data'):
            self.assertGreaterEqual(len(fig.data), 2)
            print('  plotly traces:', len(fig.data))


if __name__ == '__main__':
    unittest.main()



# coding=utf-8
# ======================================
# Package:  qteasy
# Author:   Jackie PENG
# Contact:  jackie.pengzhao@gmail.com
# Created:  2020-02-11
# Desc:
# QTEASY:
#  A fast and easy-to-use quant-investment
#  strategy research tool kit.
# ======================================

import os
import sys
import warnings
import logging
from datetime import datetime, timedelta
from logging.handlers import TimedRotatingFileHandler
from argparse import Namespace
from typing import Optional

import tushare as ts
import numpy as np

from qteasy.trade_recording import delete_account
from qteasy.qt_operator import Operator
from qteasy.visual import candle
from qteasy.parameter import Parameter

from qteasy.core import (
    run,
    info,
    is_ready,
    configure,
    get_basic_info,
    get_stock_info,
    get_data_overview,
    refill_data_source,
    get_history_data,
    get_kline,
    filter_stock_codes,
    filter_stocks,
    get_table_info,
    get_table_overview,
    live_trade_accounts,
)

from qteasy.utilfuncs import (
    is_float_like,
    is_integer_like,
)
from qteasy.configure import (
    set_config,
    configure,
    get_configurations,
    get_config,
    view_config_files,
    get_start_up_settings,
    configuration,
    save_config,
    load_config,
    reset_config,
    _parse_start_up_config_lines,
    start_up_settings,
    update_start_up_setting,
    remove_start_up_setting,
    _read_start_up_file,
)

from qteasy.history import (
    HistoryPanel,
    dataframe_to_hp,
    stack_dataframes,
)

from qteasy.strategy import (
    BaseStrategy,
    RuleIterator,
    GeneralStg,
    FactorSorter,
)

from qteasy.built_in import (
    built_ins,
    built_in_list,
    built_in_strategies,
    get_built_in_strategy,
    built_in_doc,
)


from qteasy.finance import (
    CashPlan,
    set_cost,
    update_cost,
)

from qteasy.database import (
    DataSource,
    # find_history_data,
)

from qteasy.datatypes import (
    find_history_data,
    DataType,
    StgData,
)

from qteasy._arg_validators import (
    QT_CONFIG,
    ConfigDict,
)


# qteasy版本信息
__version__ = '2.2.0'
version_info = Namespace(
        major=2,
        minor=2,
        patch=0,
        short=(2, 2),
        full=(2, 2, 0),
        string='2.2.0',
        tuple=('2', '2', '0'),
        releaselevel='beta',
)

# 解析qteasy的本地安装路径
QT_ROOT_PATH = os.path.normpath(os.path.join(os.path.dirname(__file__), os.pardir))
QT_ROOT_PATH = os.path.join(QT_ROOT_PATH, 'qteasy/')

QT_SYS_LOG_PATH = ''
QT_TRADE_LOG_PATH = ''

# 准备从本地配置文件中读取预先存储的qteasy配置
_qt_local_configs = {}

# 读取start up configurations文件内容到config_lines列表中，如果文件不存在，则创建一个空文本文件
_config_lines = _read_start_up_file(os.path.join(QT_ROOT_PATH, 'qteasy.cfg'))

# 解析config_lines列表，依次读取所有存储的属性，所有属性存储的方式为：
start_up_config = _parse_start_up_config_lines(config_lines=_config_lines)

_qt_local_configs.update(start_up_config)

# 读取tushare token，如果读取失败，抛出warning
try:
    TUSHARE_TOKEN = _qt_local_configs['tushare_token']
    ts.set_token(TUSHARE_TOKEN)
except Exception as e:
    msg = f'Failed Loading tushare_token, configure it in qteasy.cfg:\n' \
          f'tushare_token = your_token\n' \
          f'for more information, check qteasy tutorial: ' \
          f'https://qteasy.readthedocs.io/zh-cn/latest/tutorials/1-get-started.html'
    warnings.warn(msg, stacklevel=2)

# 读取其他本地配置属性，更新QT_CONFIG, 允许用户自定义参数存在
set_config(only_built_in_keys=False, **_qt_local_configs)

# 连接默认的本地数据源
QT_DATA_SOURCE = DataSource(
        source_type=QT_CONFIG['local_data_source'],
        file_type=QT_CONFIG['local_data_file_type'],
        file_loc=QT_CONFIG['local_data_file_path'],
        host=QT_CONFIG['local_db_host'],
        port=QT_CONFIG['local_db_port'],
        user=QT_CONFIG['local_db_user'],
        password=QT_CONFIG['local_db_password'],
        db_name=QT_CONFIG['local_db_name']
)

# 初始化默认交易日历
QT_TRADE_CALENDAR = QT_DATA_SOURCE.read_table_data('trade_calendar')
if not QT_TRADE_CALENDAR.empty:
    QT_TRADE_CALENDAR = QT_TRADE_CALENDAR
else:
    QT_TRADE_CALENDAR = None
    msg = 'trade calendar is not loaded, some utility functions may not work properly, ' \
            'to download trade calendar, run \n"qt.refill_data_source(tables=\'trade_calendar\')"'
    warnings.warn(msg, stacklevel=2)

# 设置qteasy运行过程中忽略某些numpy计算错误报警
np.seterr(divide='ignore', invalid='ignore')


def _validate_path_string(path_setting: str, config_key: str) -> None:
    """ 校验路径配置字符串中是否包含非法字符，若存在则抛出 ValueError。

    通用规则：禁止 ASCII 控制字符（0x00-0x1F、0x7F）。
    Windows 下额外禁止 <>"|?*，且仅允许在第二位出现冒号（如 C:）。
    """
    if not isinstance(path_setting, str):
        raise ValueError(
            f'Path configuration "{config_key}" must be a string, got {type(path_setting).__name__}.'
        )
    for c in path_setting:
        if ord(c) < 32 or ord(c) == 127:
            raise ValueError(
                f'Invalid control character in configuration "{config_key}": '
                'path must not contain control characters.'
            )
    if os.name == 'nt':
        forbidden = set('<>"|?*')
        for i, c in enumerate(path_setting):
            if c in forbidden:
                raise ValueError(
                    f'Invalid character in configuration "{config_key}": "{c}" is not allowed in path.'
                )
            if c == ':':
                if not (i == 1 and len(path_setting) >= 2 and path_setting[0].isalpha()):
                    raise ValueError(
                        f'Invalid character in configuration "{config_key}": '
                        '":" only allowed as drive letter (e.g. C:).'
                    )


def _resolve_path(root: str, path_setting: str, config_key: str = 'path') -> str:
    """ 根据配置项解析为绝对路径：支持相对路径、绝对路径与 ~ 家目录路径。

    先做非法字符校验，再 expanduser，若为绝对路径则直接 normpath，否则相对 root 拼接。
    """
    _validate_path_string(path_setting, config_key)
    expanded = os.path.expanduser(path_setting)
    if os.path.isabs(expanded):
        return os.path.normpath(expanded)
    return os.path.normpath(os.path.join(root, expanded))


def _refresh_log_paths() -> None:
    """ 根据当前 QT_CONFIG 中的 sys_log_file_path、trade_log_file_path 刷新日志路径并创建目录。

    仅对上述两个配置项生效，不修改 local_data_file_path 或 QT_DATA_SOURCE。
    """
    global QT_SYS_LOG_PATH, QT_TRADE_LOG_PATH
    QT_SYS_LOG_PATH = _resolve_path(
        QT_ROOT_PATH, QT_CONFIG['sys_log_file_path'], 'sys_log_file_path'
    )
    QT_TRADE_LOG_PATH = _resolve_path(
        QT_ROOT_PATH, QT_CONFIG['trade_log_file_path'], 'trade_log_file_path'
    )
    os.makedirs(QT_SYS_LOG_PATH, exist_ok=True)
    os.makedirs(QT_TRADE_LOG_PATH, exist_ok=True)


def _rotate_trade_logs(base_path: str, keep_days: int) -> list[str]:
    """根据保留天数删除指定目录下的旧 trade_log / trade_summary 文件。

    优先从文件名中解析创建时间，格式参考 qt_operator 中的生成逻辑：
    trade_log_{operator_name}_%Y%m%d_%H%M%S.csv
    trade_summary_{operator_name}_%Y%m%d_%H%M%S.csv
    若解析失败，则退回使用文件修改时间近似作为创建时间。
    """
    from qteasy import QT_CONFIG  # 局部导入以避免循环引用

    if keep_days is None or keep_days <= 0:
        return []

    if not os.path.isdir(base_path):
        return []

    threshold = datetime.now() - timedelta(days=keep_days)
    removed_files: list[str] = []

    for name in os.listdir(base_path):
        full_path = os.path.join(base_path, name)
        if not os.path.isfile(full_path):
            continue
        if not (name.startswith('trade_log_') or name.startswith('trade_summary_') or name.startswith('value_curve_')):
            continue
        if not name.endswith('.csv'):
            continue

        created_time: Optional[datetime] = None

        try:
            # 解析文件名中的时间戳部分：{prefix}_{operator_name}_%Y%m%d_%H%M%S.csv
            stem = name[:-4]  # 去除 .csv
            parts = stem.split('_')
            if len(parts) >= 3:
                date_str, time_str = parts[-2], parts[-1]
                created_time = datetime.strptime(f'{date_str}_{time_str}', '%Y%m%d_%H%M%S')
        except Exception:
            created_time = None

        if created_time is None:
            # 解析失败时退回到 mtime
            try:
                created_time = datetime.fromtimestamp(os.path.getmtime(full_path))
            except Exception:
                continue

        if created_time < threshold:
            try:
                os.remove(full_path)
                removed_files.append(full_path)
            except Exception as e:
                # 仅记录 warning，不中断程序
                logging.getLogger('core').warning(
                    'Failed to remove old trade log file "%s": %s', full_path, e
                )

    if removed_files:
        logging.getLogger('core').info(
            'Rotated trade logs: %d files removed, keep_days=%s',
            len(removed_files),
            keep_days,
        )

    return removed_files


def _auto_rotate_trade_logs() -> None:
    """在模块初始化阶段，根据配置自动对当前 trade_log 目录做一次轮换删除。"""
    global QT_TRADE_LOG_PATH

    keep_days = QT_CONFIG.get('trade_log_keep_days', 0)
    if keep_days is None or (isinstance(keep_days, int) and keep_days <= 0):
        return

    _rotate_trade_logs(QT_TRADE_LOG_PATH, int(keep_days))


def rotate_trade_logs(days: Optional[int] = None) -> None:
    """手动触发 trade_log / trade_summary 轮换删除。

    Parameters
    ----------
    days : int or None
        若为 None，则使用全局配置 trade_log_keep_days；
        若为正整数，则以该值作为本次删除的保留天数（不修改全局配置）。
    """
    global QT_TRADE_LOG_PATH

    if days is None:
        keep_days = QT_CONFIG.get('trade_log_keep_days', 0)
    else:
        if not isinstance(days, int):
            raise ValueError(f'Parameter "days" must be an integer or None, got {type(days).__name__}.')
        keep_days = days

    if keep_days is None or (isinstance(keep_days, int) and keep_days <= 0):
        return

    _rotate_trade_logs(QT_TRADE_LOG_PATH, int(keep_days))


# 设置qteasy回测交易报告以及错误报告的存储路径（支持热修改，见 configure 挂钩）
_refresh_log_paths()
# 设置loggings，创建logger
debug_handler = logging.handlers.TimedRotatingFileHandler(filename=os.path.join(QT_SYS_LOG_PATH, 'qteasy.log'),
                                                          backupCount=3, when='midnight')
error_handler = logging.StreamHandler()
debug_handler.setLevel(logging.DEBUG)
error_handler.setLevel(logging.ERROR)
formatter = logging.Formatter('[%(asctime)s]:%(levelname)s - %(module)s -: %(message)s')
debug_handler.setFormatter(formatter)
error_handler.setFormatter(formatter)
# 创建两个logger，一个是正常运行(回测、优化等）的logger，另一个是实盘运行logger
logger_core = logging.getLogger('core')
logger_core.addHandler(debug_handler)
logger_core.addHandler(error_handler)
# 根据 QT_CONFIG['log_level'] 设置日志等级，默认 INFO
_log_level_name = str(QT_CONFIG.get('log_level', 'INFO')).upper()
_log_level_value = getattr(logging, _log_level_name, logging.INFO)
logger_core.setLevel(_log_level_value)

# 在 logger 初始化完成后，根据配置自动做一次 trade_log 轮换
_auto_rotate_trade_logs()
logger_core.propagate = False

logger_core.info('qteasy loaded!')

# 检查当前python的版本信息，打印相关的warning信息
py_version = sys.version_info
py_ver_major = py_version.major
py_ver_minor = py_version.minor

# qteasy的全局常量
# 运行模式常量
LIVE_TRADE_MODE = 0
LIVE_MODE = 0
BACKTEST_MODE = 1
OPTIMIZE_MODE = 2
OPTI_MODE = 2
OPTIMIZATION_MODE = 2
PREDICT_MODE = 3
PREDICTION_MODE = 3

__all__ = [
    'run', 'set_config', 'get_configurations', 'get_config', 'view_config_files',
    'info', 'is_ready', 'configure', 'configuration', 'save_config', 'load_config', 'reset_config',
    'get_basic_info', 'get_stock_info', 'get_data_overview', 'refill_data_source',
    'get_history_data', 'filter_stock_codes', 'filter_stocks', 'start_up_config', 'Parameter',
    'get_table_info', 'get_table_overview', 'get_start_up_settings', 'find_history_data', 'DataType', 'StgData',
    'HistoryPanel', 'dataframe_to_hp', 'stack_dataframes', 'start_up_settings', 'update_start_up_setting', 'get_kline',
    'Operator', 'BaseStrategy', 'RuleIterator', 'GeneralStg', 'FactorSorter', 'remove_start_up_setting',
    'built_ins', 'built_in_list', 'built_in_strategies', 'get_built_in_strategy',
    'candle', 'CashPlan', 'set_cost', 'update_cost', 'DataSource', 'find_history_data',
    'QT_TRADE_CALENDAR', 'QT_TRADE_LOG_PATH', 'QT_ROOT_PATH', 'QT_SYS_LOG_PATH',
    'QT_DATA_SOURCE', 'QT_CONFIG', 'utilfuncs', 'QT_CONFIG', 'ConfigDict', '__version__', 'version_info',
    'logger_core', 'live_trade_accounts', 'delete_account', 'LIVE_TRADE_MODE', 'LIVE_MODE', 'BACKTEST_MODE',
    'OPTIMIZE_MODE', 'OPTI_MODE', 'OPTIMIZATION_MODE', 'PREDICT_MODE', 'PREDICTION_MODE',
]
